<?php
    # PHP Weathermap v0.82 - Pre-release software. For testing
    # Copyright Howard Jones, 2005-2006 howie@thingy.com
    # http://wotsit.thingy.com/haj/cacti/
    # Released under the GNU Public License

    require_once "HTML_ImageMap.class.php";

    $WEATHERMAP_VERSION="0.82";
    $weathermap_debugging=FALSE;

    // Turn on ALL error reporting for now.
    error_reporting (E_ALL);

    // Utility functions
    // Check for GD & PNG support This is just in here so that both the editor and CLI can use it without the need for another file
    function module_checks()
    {
        if (!extension_loaded('gd'))
        {
            warn ("\n\nNo image (gd) extension is loaded. This is required by weathermap.\n\n");
            warn ("\nIMPORTANT: Just because mod_php and php_info() shows gd in a webpage, it doesn't mean it's\nAvailable in your command-line php program. Please check the output from 'php -i' and your 'php.ini' to be sure!\n\n");
            
            return (FALSE);
        }

        if (!function_exists('imagecreatefrompng'))
        {
            warn ("Your GD php module doesn't support PNG format.\n");
            return (FALSE);
        }
        
        if (!function_exists('imagecreatetruecolor'))
        {
            warn ("Your GD php module doesn't support truecolor.\n");
            return (FALSE);
        }
        
        if (!function_exists('imagecopyresampled'))
        {
            warn ("Your GD php module doesn't support thumbnail creation (imagecopyresampled).\n");
        }
        return (TRUE);
    }

    function debug($string)
    {
        global $weathermap_debugging;

        if ($weathermap_debugging)
        {
            // use Cacti's debug log, if we are running from the poller
            if (function_exists('debug_log_insert') && (!function_exists('show_editor_startpage')))
                { cacti_log(rtrim($string), true, "WEATHERMAP"); }
            else
            {
                $stderr=fopen('php://stderr', 'w');
                fwrite($stderr, $string);
                fclose ($stderr);
            }
        }
    }

    function warn($string)
    {
        // use Cacti's debug log, if we are running from the poller
        if (function_exists('cacti_log') && (!function_exists('show_editor_startpage')))
            { cacti_log(rtrim($string), true, "WEATHERMAP"); }
        else
        {
            $stderr=fopen('php://stderr', 'w');
            fwrite($stderr, $string);
            fclose ($stderr);
        }
    }

    function js_escape($str)
    {
        $str=str_replace('\\', '\\\\', $str);
        $str=str_replace("'", "\\\'", $str);

        $str="'" . $str . "'";

        return ($str);
    }

    // wrapper around imagecolorallocate to try and re-use palette slots where possible
    function myimagecolorallocate($image, $red, $green, $blue)
    {
        $existing=imagecolorexact($image, $red, $green, $blue);

        if ($existing > -1)
            return $existing;

        return (imagecolorallocate($image, $red, $green, $blue));
    }

    function render_colour($col)
    {
        if (($col[0] < 0) && ($col[1] < 0) && ($col[1] < 0)) { return 'none'; }
        else { return sprintf("%d %d %d", $col[0], $col[1], $col[2]); }
    }

    // take the same set of points that imagepolygon does, but don't close the shape
    function imagepolyline($image, $points, $npoints, $color)
    {
        for ($i=0; $i < ($npoints - 1);
            $i++) { imageline($image, $points[$i * 2], $points[$i * 2 + 1], $points[$i * 2 + 2], $points[$i * 2 + 3],
                              $color); }
    }

    // calculate the points for a span of the curve. We pass in the distance so far, and the array index, so that
    // the chunk of array generated by this function can be array_merged with existing points from before.
    // Considering how many array functions there are, PHP has horrible list support
    // Each point is a 3-tuple - x,y,distance - which is used later to figure out where the 25%, 50% marks are on the curve
    function calculate_catmull_rom_span($startn, $startdistance, $numsteps, $x0, $y0, $x1, $y1, $x2, $y2, $x3, $y3)
    {
        $Ap_x=-$x0 + 3 * $x1 - 3 * $x2 + $x3;
        $Bp_x=2 * $x0 - 5 * $x1 + 4 * $x2 - $x3;
        $Cp_x=-$x0 + $x2;
        $Dp_x=2 * $x1;

        $Ap_y=-$y0 + 3 * $y1 - 3 * $y2 + $y3;
        $Bp_y=2 * $y0 - 5 * $y1 + 4 * $y2 - $y3;
        $Cp_y=-$y0 + $y2;
        $Dp_y=2 * $y1;

        $d=2;
        $n=$startn;
        $distance=$startdistance;

        $lx=$x0;
        $ly=$y0;

        $allpoints[]=array
            (
                $x0,
                $y0,
                $distance
            );

        for ($i=0; $i <= $numsteps; $i++)
        {
            $t=$i / $numsteps;
            $t2=$t * $t;
            $t3=$t2 * $t;
            $x=(($Ap_x * $t3) + ($Bp_x * $t2) + ($Cp_x * $t) + $Dp_x) / $d;
            $y=(($Ap_y * $t3) + ($Bp_y * $t2) + ($Cp_y * $t) + $Dp_y) / $d;

            if ($i > 0)
            {
                $step=sqrt((($x - $lx) * ($x - $lx)) + (($y - $ly) * ($y - $ly)));
                $distance=$distance + $step;
                $allpoints[$n]=array
                    (
                        $x,
                        $y,
                        $distance
                    );

                $n++;
            }

            $lx=$x;
            $ly=$y;
        }

        return array($allpoints, $distance, $n);
    }

    // return the index of the point either at (unlikely) or just before the target distance
    // we will linearly interpolate afterwards to get a true point - pointarray is an array of 3-tuples produced by the function above
    function find_distance($pointarray, $distance)
    {
        $left=0;
        $right=count($pointarray) - 1;

        if ($left == $right)
            return ($left);

        // if it's a point past the end of the line, then just return the end of the line
        // Weathermap should *never* ask for this, anyway
        if ($pointarray[$right][2] < $distance) { return ($right); }

        // if somehow we have a 0-length curve, then don't try and search, just give up
        // in a somewhat predictable manner
        if ($pointarray[$left][2] == $pointarray[$right][2]) { return ($left); }

        while ($left <= $right)
        {
            $mid=floor(($left + $right) / 2);

            if (($pointarray[$mid][2] < $distance) && ($pointarray[$mid + 1][2] >= $distance)) { return $mid; }

            if ($distance <= $pointarray[$mid][2]) { $right=$mid - 1; }
            else { $left=$mid + 1; }
        }

        print "FELL THROUGH\n";
        die ("Howie's crappy binary search is wrong after all.\n");
    }

    // top-level function that takes a two lists to define some points, and draws a weathermap link
    // - this takes care of all the extras, like arrowheads, and where to put the bandwidth labels
    //    xarray and yarray are the points the curve passes through
    //    width is the link width (the actual width is twice this)
    //    outlinecolour is a GD colour reference
    //    fillcolours is an array of two more colour references, one for the out, and one for the in spans
    //    pointsperspan is optional, and intended to allow fine-tuning of the smoothness of curves. 15 seems about right for most cases
    //   -- returns an array of arrays for the two quarter-points - this is where the
    //       bandwidth labels go, if they were required
    function draw_curve($image, $in_xarray, $in_yarray, $width, $outlinecolour, $fillcolours, $linkname, &$map,
                        $pointsperspan = 12)
    {
        global $marker2, $marker3;

        # search through the point list, for consecutive duplicate points
        # (most common case will be a straight link with both NODEs at the same place, I think)
        # strip those out, because they'll break the binary search/centre-point stuff

        $last_x=NULL;
        $last_y=NULL;

        for ($i=0; $i < count($in_xarray); $i++)
        {
            if (($in_xarray[$i] == $last_x) && ($in_yarray[$i] == $last_y)) { debug
                ("Dumping useless duplicate point on curve\n"); }
            else
            {
                $xarray[]=$in_xarray[$i];
                $yarray[]=$in_yarray[$i];
            }

            $last_x=$in_xarray[$i];
            $last_y=$in_yarray[$i];
        }

        // only proceed if we still have at least two points!
        if (count($xarray) > 1)
        {

            # duplicate the first and last points, so that all points are drawn
            # (C-R normally would draw from x[1] to x[n-1]
            array_unshift($xarray, $xarray[0]);
            array_unshift($yarray, $yarray[0]);

            $x=array_pop($xarray);
            $y=array_pop($yarray);
            array_push($xarray, $x);
            array_push($xarray, $x);
            array_push($yarray, $y);
            array_push($yarray, $y);

            $npoints=count($xarray);

            $curvepoints=array
                (
                );

            // add in the very first point manually (the calc function skips this one to avoid duplicates, which mess up the distance stuff)
            $curvepoints[]=array
                (
                    $xarray[0],
                    $yarray[0],
                    0
                );

            $startindex=0;
            $startdistance=0;

            for ($i=0; $i < ($npoints - 3); $i++)
            {
                list($newpoints,
                     $distance,
                     $np)=calculate_catmull_rom_span($startindex,     $startdistance,  $pointsperspan,  $xarray[$i],
                                                     $yarray[$i],     $xarray[$i + 1], $yarray[$i + 1], $xarray[$i + 2],
                                                     $yarray[$i + 2], $xarray[$i + 3], $yarray[$i + 3]);
                $curvepoints=$curvepoints + $newpoints;

                $startindex=$np;
                $startdistance=$distance;
            }

            // now we have a 'spine' - all the central points for this curve.
            // time to flesh it out to the right width, and figure out where to draw arrows and bandwidth boxes...

            $totaldistance=$startdistance;
            $halfway=$totaldistance / 2;
            $q1=$totaldistance / 4;
            $q3=$halfway + $q1;

            // We find the nearest lower point for each distance, then linearly interpolate to get a more acccurate point
            // this saves having quite so many points-per-curve

            $halfwayindex=find_distance($curvepoints, $halfway);

            $ratio=($halfway - $curvepoints[$halfwayindex][2])
                       / ($curvepoints[$halfwayindex + 1][2] - $curvepoints[$halfwayindex][2]);
            $halfway_x=$curvepoints[$halfwayindex][0] + $ratio
                           * ($curvepoints[$halfwayindex + 1][0] - $curvepoints[$halfwayindex][0]);
            $halfway_y=$curvepoints[$halfwayindex][1] + $ratio
                           * ($curvepoints[$halfwayindex + 1][1] - $curvepoints[$halfwayindex][1]);

            $arrowlengthfactor=4;
            $arrowwidthfactor=2;

            if ($map->links[$linkname]->arrowstyle == 'compact')
            {
                $arrowlengthfactor=1;
                $arrowwidthfactor=1;
            }

            if (preg_match('/(\d+) (\d+)/', $map->links[$linkname]->arrowstyle, $matches))
            {
                $arrowlengthfactor=$matches[1];
                $arrowwidthfactor=$matches[2];
            }

            $arrowsize=$width * $arrowlengthfactor;
            $arrowwidth=$width * $arrowwidthfactor;

            // loop over direction here
            // direction is 1.0 for the first half (forwards through the pointlist), and -1.0 for the second half (backwards from the end)
            //  - used as a multiplier on anything that looks forwards or backwards through the list

            foreach (array(1.0, -1.0)as $direction)
            {
                // this is the last index before the arrowhead starts
                $pre_midindex=find_distance($curvepoints, $halfway - $direction * $arrowsize);

                $ratio=(($halfway - $direction * $arrowsize) - $curvepoints[$pre_midindex][2])
                           / ($curvepoints[$pre_midindex + $direction][2] - $curvepoints[$pre_midindex][2]);
                $pre_mid_x=$curvepoints[$pre_midindex][0] + $ratio
                               * ($curvepoints[$pre_midindex + $direction][0] - $curvepoints[$pre_midindex][0]);
                $pre_mid_y=$curvepoints[$pre_midindex][1] + $ratio
                               * ($curvepoints[$pre_midindex + $direction][1] - $curvepoints[$pre_midindex][1]);

                $there_points=array
                    (
                    );

                $back_points=array
                    (
                    );

                $arrowpoints=array
                    (
                    );

                if ($direction < 0) { $start=count($curvepoints) - 1; }
                else { $start=0; }

                for ($i=$start; $i != $pre_midindex; $i+=$direction)
                {
                    // for each point on the spine, produce two points normal to it's direction,
                    // each is $width away from the spine, but we build up the two lists in the opposite order,
                    // so that when they are joined together, we get one continuous line

                    $dx=$curvepoints[$i + $direction][0] - $curvepoints[$i][0];
                    $dy=$curvepoints[$i + $direction][1] - $curvepoints[$i][1];
                    $l=sqrt(($dx * $dx) + ($dy * $dy));
                    $nx=$dy / $l;
                    $ny=-$dx / $l;

                    $there_points[]=$curvepoints[$i][0] + $direction * $width * $nx;
                    $there_points[]=$curvepoints[$i][1] + $direction * $width * $ny;

                    $back_points[]=$curvepoints[$i][0] - $direction * $width * $nx;
                    $back_points[]=$curvepoints[$i][1] - $direction * $width * $ny;
                }

                // all the normal line is done, now lets add an arrowhead on

                $adx=($halfway_x - $pre_mid_x);
                $ady=($halfway_y - $pre_mid_y);
                $l=sqrt(($adx * $adx) + ($ady * $ady));

                $anx=$ady / $l;
                $any=-$adx / $l;

                $there_points[]=$pre_mid_x + $direction * $width * $anx;
                $there_points[]=$pre_mid_y + $direction * $width * $any;

                $there_points[]=$pre_mid_x + $direction * $arrowwidth * $anx;
                $there_points[]=$pre_mid_y + $direction * $arrowwidth * $any;

                $there_points[]=$halfway_x;
                $there_points[]=$halfway_y;

                $there_points[]=$pre_mid_x - $direction * $arrowwidth * $anx;
                $there_points[]=$pre_mid_y - $direction * $arrowwidth * $any;

                $there_points[]=$pre_mid_x - $direction * $width * $anx;
                $there_points[]=$pre_mid_y - $direction * $width * $any;

                // all points done, now combine the lists, and produce the final result.

                $y=array_pop($back_points);
                $x=array_pop($back_points);

                do
                {
                    $there_points[]=$x;
                    $there_points[]=$y;
                    $y=array_pop($back_points);
                    $x=array_pop($back_points);
                } while (!is_null($y));

                $arrayindex=0;

                if ($direction < 0)
                    $arrayindex=1;

                if (!is_null($fillcolours[$arrayindex]))
                    { imagefilledpolygon($image, $there_points, count($there_points) / 2, $fillcolours[$arrayindex]); }

                $map->imap->addArea("Polygon", "LINK:" . $linkname, '', $there_points);
                debug ("Adding Poly imagemap for $linkname\n");

                if (!is_null($outlinecolour))
                    imagepolygon($image, $there_points, count($there_points) / 2, $outlinecolour);
            }

            // now figure out where the labels should go, so we can pass those
            // coordinates back

            $q1index=find_distance($curvepoints, $q1);
            $ratio=($q1 - $curvepoints[$q1index][2]) / ($curvepoints[$q1index + 1][2] - $curvepoints[$q1index][2]);
            $q1_x=$curvepoints[$q1index][0] + $ratio * ($curvepoints[$q1index + 1][0] - $curvepoints[$q1index][0]);
            $q1_y=$curvepoints[$q1index][1] + $ratio * ($curvepoints[$q1index + 1][1] - $curvepoints[$q1index][1]);

            $q3index=find_distance($curvepoints, $q3);
            $ratio=($q3 - $curvepoints[$q3index][2]) / ($curvepoints[$q3index + 1][2] - $curvepoints[$q3index][2]);
            $q3_x=$curvepoints[$q3index][0] + $ratio * ($curvepoints[$q3index + 1][0] - $curvepoints[$q3index][0]);
            $q3_y=$curvepoints[$q3index][1] + $ratio * ($curvepoints[$q3index + 1][1] - $curvepoints[$q3index][1]);

            return (array($q1_x, $q1_y, $q3_x, $q3_y));
        }
        else
        {
            warn ("Arrow not drawn, as it's 1-dimensional.\n");
            return (array(NULL, NULL, NULL, NULL));
        }
    }

    function unformat_number($instring, $kilo = 1000)
    {
        $matches=0;
        $number=0;

        if (preg_match("/([0-9\.]+)(M|G|K|T)/", $instring, $matches))
        {
            $number=floatval($matches[1]);

            if ($matches[2] == 'K') { $number=$number * $kilo; }

            if ($matches[2] == 'M') { $number=$number * $kilo * $kilo; }

            if ($matches[2] == 'G') { $number=$number * $kilo * $kilo * $kilo; }

            if ($matches[2] == 'T') { $number=$number * $kilo * $kilo * $kilo * $kilo; }

            // new, for absolute datastyle. Think seconds.
            if ($matches[2] == 'm') { $number=$number / $kilo; }

            if ($matches[2] == 'u') { $number=$number / ($kilo * $kilo); }
        }
        else { $number=floatval($instring); }

        return ($number);
    }

    // given a compass-point, and a width & height, return a tuple of the x,y offsets
    function calc_offset($offsetstring, $width, $height)
    {
        switch (strtoupper($offsetstring))
        {
            case 'N':
                return (array(0, -$height / 2));

                break;

            case 'S':
                return (array(0, $height / 2));

                break;

            case 'E':
                return (array(+$width / 2, 0));

                break;

            case 'W':
                return (array(-$width / 2, 0));

                break;

            case 'NW':
                return (array(-$width / 2, -$height / 2));

                break;

            case 'NE':
                return (array($width / 2, -$height / 2));

                break;

            case 'SW':
                return (array(-$width / 2, $height / 2));

                break;

            case 'SE':
                return (array($width / 2, $height / 2));

                break;

            default:
                return (array(0, 0));

                break;
        }
    }

    // These next two are based on perl's Number::Format module
    // by William R. Ward, chopped down to just what I needed

    function format_number($number, $precision = 2, $trailing_zeroes = 0)
    {
        $sign=1;

        if ($number < 0)
        {
            $number=abs($number);
            $sign=-1;
        }

        $number=round($number, $precision);
        $integer=intval($number);

        if (strlen($integer) < strlen($number)) { $decimal=substr($number, strlen($integer) + 1); }

        if (!isset($decimal)) { $decimal=''; }

        $integer=$sign * $integer;

        if ($decimal == '') { return ($integer); }
        else { return ($integer . "." . $decimal); }
    }

    function nice_bandwidth($number, $kilo = 1000)
    {
        $suffix='';

        if ($number == 0)
            return '0';

        $mega=$kilo * $kilo;
        $giga=$mega * $kilo;
        $tera=$giga * $kilo;

        if ($number > $tera)
        {
            $number/=$tera;
            $suffix="T";
        }
        elseif ($number > $giga)
        {
            $number/=$giga;
            $suffix="G";
        }
        elseif ($number > $mega)
        {
            $number/=$mega;
            $suffix="M";
        }
        elseif ($number > $kilo)
        {
            $number/=$kilo;
            $suffix="K";
        }

        $result=format_number($number, 1) . $suffix;
        return ($result);
    }

    function nice_scalar($number, $kilo = 1000)
    {
        $suffix='';

        if ($number == 0)
            return '0';

        $mega=$kilo * $kilo;
        $giga=$mega * $kilo;
        $tera=$giga * $kilo;

        if ($number > $tera)
        {
            $number/=$tera;
            $suffix="T";
        }
        elseif ($number > $giga)
        {
            $number/=$giga;
            $suffix="G";
        }
        elseif ($number > $mega)
        {
            $number/=$mega;
            $suffix="M";
        }
        elseif ($number > $kilo)
        {
            $number/=$kilo;
            $suffix="K";
        }
        elseif ($number < (1 / ($kilo)))
        {
            $number=$number * $mega;
            $suffix="u";
        }
        elseif ($number < 1)
        {
            $number=$number * $kilo;
            $suffix="m";
        }

        $result=format_number($number, 1) . $suffix;
        return ($result);
    }

    // ***********************************************

    class WeatherMapNode
        {
            var $owner;
            var $x,
                $y;
            var $width,
                $height;
            var $label,
                $labelfont;
            var $name;
            var $infourl;
            var $overliburl;
            var $overlibwidth,
                $overlibheight;
            var $maphtml;
            var $selected = 0;
            var $iconfile;
            var $labeloffset,
                $labeloffsetx,
                $labeloffsety;

            var $inherit_fieldlist;

            var $labelbgcolour;
            var $labeloutlinecolour;
            var $labelfontcolour;
            var $labelfontshadowcolour;

            function WeatherMapNode()
            {
                $this->inherit_fieldlist=array
                    (
                        'label' => '',
                        'labelfont' => 3,
                        'x' => 0,
                        'y' => 0,
                        'iconfile' => '',
                        'infourl' => '',
                        'overliburl' => '',
                        'overlibwidth' => 0,
                        'overlibheight' => 0,
                        'labeloutlinecolour' => array(0, 0, 0),
                        'labelbgcolour' => array(255, 255, 255),
                        'labelfontcolour' => array(0, 0, 0),
                        'labelfontshadowcolour' => array(-1, -1, -1),
                        'labeloffset' => '',
                        'labeloffsetx' => 0,
                        'labeloffsety' => 0
                    );

                $this->width=0;
                $this->height=0;
            }

            function calc_size()
            {
                $this->width=0;
                $this->height=0;

                // calculate the size of the NODE box, so we can make links end at corners.
                if ($this->label != '')
                {
                    $padding=0;
                    $font=$this->labelfont;

                    list($strwidth, $strheight)=$this->owner->myimagestringsize($font, $this->label);

                    $boxwidth=$strwidth * 1.1;
                    $boxheight=$strheight * 1.1;

                    $this->width=$boxwidth;
                    $this->height=$boxheight;
                }

                // if there's an icon, then that's what the corners relate to
                if ($this->iconfile != '')
                {
                    $temp_im=imagecreatefrompng($this->iconfile);

                    if ($temp_im)
                    {
                        $this->width=imagesx($temp_im);
                        $this->height=imagesy($temp_im);
                    }

                    imagedestroy ($temp_im);
                }

                debug ("PRECALC $this->name: $this->width x $this->height\n");
            }

            function Reset($newowner)
            {
                $this->owner=$newowner;

                if (isset($this->owner->defaultnode) && $this->name != 'DEFAULT') {
                // use the defaults from DEFAULT
                    $this->CopyFrom($this->owner->defaultnode); }
                else
                {
                    # use the default defaults
                    foreach (array_keys($this->inherit_fieldlist)as
                        $fld) { $this->$fld=$this->inherit_fieldlist[$fld]; }
                }
            }

            function CopyFrom($source)
            {
                foreach (array_keys($this->inherit_fieldlist)as $fld) { $this->$fld=$source->$fld; }
            }

            function WriteConfig($fd)
            {
                $output='';

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['label'] : $this->owner->defaultnode->label);

                if ($this->label != $comparison) { $output.="\tLABEL " . $this->label . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['infourl'] : $this->owner->defaultnode->infourl);

                if ($this->infourl != $comparison) { $output.="\tINFOURL " . $this->infourl . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['overliburl'] : $this->owner->defaultnode->overliburl);

                if ($this->overliburl != $comparison) { $output.="\tOVERLIBGRAPH " . $this->overliburl . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['iconfile'] : $this->owner->defaultnode->iconfile);

                if ($this->iconfile != $comparison) { $output.="\tICON " . $this->iconfile . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['labelfont'] : $this->owner->defaultnode->labelfont);

                if ($this->labelfont != $comparison) { $output.="\tLABELFONT " . $this->labelfont . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['labeloffset'] : $this->owner->defaultnode->labeloffset);

                if ($this->labeloffset != $comparison) { $output.="\tLABELOFFSET " . $this->labeloffset . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['overlibwidth'] : $this->owner->defaultnode->overlibwidth);

                if ($this->overlibwidth != $comparison) { $output.="\tOVERLIBWIDTH " . $this->overlibwidth . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['overlibheight'] : $this->owner->defaultnode->overlibheight);

                if ($this->overlibheight != $comparison) { $output.="\tOVERLIBHEIGHT " . $this->overlibheight . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['labelbgcolour'] : $this->owner->defaultnode->labelbgcolour);

                if ($this->labelbgcolour != $comparison) { $output.="\tLABELBGCOLOR " . render_colour(
                                                                                            $this->labelbgcolour)
                                                                        . "\n"; }

                $comparison=($this->name == 'DEFAULT' ? $this->inherit_fieldlist['labelfontcolour']
                                : $this->owner->defaultnode->labelfontcolour);

                if ($this->labelfontcolour != $comparison) { $output.="\tLABELFONTCOLOR " . render_colour(
                                                                                                $this->labelfontcolour)
                                                                          . "\n"; }

                $comparison=($this->name == 'DEFAULT' ? $this->inherit_fieldlist['labeloutlinecolour']
                                : $this->owner->defaultnode->labeloutlinecolour);

                if ($this->labeloutlinecolour != $comparison) { $output.="\tLABELOUTLINECOLOR " . render_colour(
                                                                                                      $this->labeloutlinecolour) . "\n";
                    }

                $comparison=($this->name == 'DEFAULT' ? $this->inherit_fieldlist['labelfontshadowcolour']
                                : $this->owner->defaultnode->labelfontshadowcolour);

                if ($this->labelfontshadowcolour != $comparison)
                    { $output.="\tLABELFONTSHADOWCOLOR " . render_colour($this->labelfontshadowcolour) . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['labeloffsetx'] : $this->owner->defaultnode->labeloffsetx);
                $comparison2=($this->name == 'DEFAULT'
                                 ? $this->inherit_fieldlist['labeloffsety'] : $this->owner->defaultnode->labeloffsety);

                if (($this->labeloffsetx != $comparison) || ($this->labeloffsety != $comparison2))
                    { $output.="\tLABELOFFSET " . $this->labeloffsetx . " " . $this->labeloffsety . "\n"; }

                $comparison=($this->name == 'DEFAULT' ? $this->inherit_fieldlist['x'] : $this->owner->defaultnode->x);
                $comparison2=($this->name == 'DEFAULT' ? $this->inherit_fieldlist['y'] : $this->owner->defaultnode->y);

                if (($this->x != $comparison) || ($this->y != $comparison2))
                    { $output.="\tPOSITION " . $this->x . " " . $this->y . "\n"; }

                if ($output != '')
                {
                    fwrite($fd, "NODE " . $this->name . "\n");
                    fwrite($fd, "$output\n");
                }
            }

            function asJS()
            {
                $js='';
                $js.="Nodes[" . js_escape($this->name) . "] = {";
                $js.="x:" . $this->x . ", ";
                $js.="y:" . $this->y . ", ";
                $js.="label:" . js_escape($this->label) . ", ";
                $js.="name:" . js_escape($this->name) . ", ";
                $js.="infourl:" . js_escape($this->infourl) . ", ";
                $js.="overliburl:" . js_escape($this->overliburl) . ", ";
                $js.="overlibwidth:" . $this->overlibheight . ", ";
                $js.="overlibheight:" . $this->overlibwidth . ", ";
                $js.="iconfile:" . js_escape($this->iconfile);
                $js.="};\n";
                return $js;
            }
            
            function asJSON()
            {
                $js = '';
                $js .= "" . js_escape($this->name) . ": {";
                $js .= "x:" . $this->x . ", ";
                $js .= "y:" . $this->y . ", ";
                $js .= "label:" . js_escape($this->label) . ", ";
                $js .= "name:" . js_escape($this->name) . ", ";
                $js .= "infourl:" . js_escape($this->infourl) . ", ";
                $js .= "overliburl:" . js_escape($this->overliburl) . ", ";
                $js .= "overlibwidth:" . $this->overlibheight . ", ";
                $js .= "overlibheight:" . $this->overlibwidth . ", ";
                $js .= "iconfile:" . js_escape($this->iconfile);
                $js .= "},\n";
                return $js;
            }

            function Draw($im, &$map)
            {
                $strwidth=0;
                $strheight=0;

                // we do this little bit first, so that the label-offset stuff can know it
                if ($this->label != '')
                {
                    $padding=0;
                    $font=$this->labelfont;

                    list($strwidth, $strheight)=$map->myimagestringsize($font, $this->label);

                    $boxwidth=$strwidth * 1.0;
                    $boxheight=$strheight * 1.0;

                    debug ("Node->Draw: Metrics are: $font $strwidth x $strheight -> $boxwidth x $boxheight\n");
                }

                if ($this->iconfile != '')
                {
                    if (is_readable($this->iconfile))
                    {
                        imagealphablending($im, true);
                        // draw the supplied icon, instead of the labelled box
                        $temp_im=@imagecreatefrompng($this->iconfile);

                        if ($temp_im)
                        {
                            $w=imagesx($temp_im);
                            $h=imagesy($temp_im);
                            $x1=$this->x - $w / 2;
                            $y1=$this->y - $h / 2;
                            $x2=$this->x + $w / 2;
                            $y2=$this->y + $h / 2;

                            imagecopy($im, $temp_im, $x1, $y1, 0, 0, $w, $h);
                            $map->imap->addArea("Rectangle", "NODE:" . $this->name, '', array($x1, $y1, $x2, $y2));
                            imagedestroy ($temp_im);

                            if ($this->labeloffset != '')
                            {
                                $this->labeloffsetx=0;
                                $this->labeloffsety=0;

                                list($dx, $dy)=calc_offset($this->labeloffset, ($w + $strwidth), ($h + $strheight));

                                $this->labeloffsetx=$dx;
                                $this->labeloffsety=$dy;
                            }
                        }
                        else { warn ("Couldn't open PNG ICON: " . $this->iconfile . " - is it a PNG?\n"); }
                    }
                    else { warn ("ICON " . $this->iconfile
                                    . " does not exist, or is not readble. Check path and permissions.\n"); }
                }

                if ($this->label != '')
                {
                    $x=$this->x + $this->labeloffsetx;
                    $y=$this->y + $this->labeloffsety;

                    $x1=$x - ($boxwidth / 2) - 2;
                    $x2=$x + ($boxwidth / 2) + 2;
                    $y1=$y - ($boxheight / 2) - 2;
                    $y2=$y + ($boxheight / 2) + 2;

                    $txt_x=$x - $strwidth / 2;
                    $txt_y=$y + $strheight / 2;

                    if ($this->iconfile == '')
                    {
                        if ($this->labelbgcolour != array
                            (
                                -1,
                                -1,
                                -1
                            ))
                        {
                            $col=myimagecolorallocate($im, $this->labelbgcolour[0], $this->labelbgcolour[1],
                                                      $this->labelbgcolour[2]);
                            imagefilledrectangle($im, $x1, $y1, $x2, $y2, $col);
                        }

                        if ($this->selected)
                        {
                            imagerectangle($im, $x1, $y1, $x2, $y2, $map->selected);
                            // would be nice if it was thicker, too...
                            imagerectangle($im, $x1 - 1, $y1 - 1, $x2 + 1, $y2 + 1, $map->selected);
                        }
                        else
                        {
                            if ($this->labeloutlinecolour != array
                                (
                                    -1,
                                    -1,
                                    -1
                                ))
                            {
                                $col=myimagecolorallocate($im,                          $this->labeloutlinecolour[0],
                                                          $this->labeloutlinecolour[1], $this->labeloutlinecolour[2]);
                                imagerectangle($im, $x1, $y1, $x2, $y2, $col);
                            }
                        }
                    }

                    if ($this->labelfontshadowcolour != array
                        (
                            -1,
                            -1,
                            -1
                        ))
                    {
                        $col=myimagecolorallocate($im, $this->labelfontshadowcolour[0], $this->labelfontshadowcolour[1],
                                                  $this->labelfontshadowcolour[2]);
                        $map->myimagestring($im, $font, $txt_x + 1, $txt_y + 1, $this->label, $col);
                    }

                    $col=myimagecolorallocate($im, $this->labelfontcolour[0], $this->labelfontcolour[1],
                                              $this->labelfontcolour[2]);
                    $map->myimagestring($im, $font, $txt_x, $txt_y, $this->label, $col);

                    $map->imap->addArea("Rectangle", "NODE:" . $this->name, '', array($x1, $y1, $x2, $y2));
                }
            }
        }

    ;

    class WeatherMapLink
        {
            var $owner,                $name;
            var $maphtml;
            var $a,                    $b; // the ends - references to nodes
            var $width,                $arrowstyle;
            var $bwfont,               $labelstyle;
            var $overliburl,           $infourl;
            var $bandwidth_in,         $bandwidth_out;
            var $overlibwidth,         $overlibheight;
            var $max_bandwidth_in,     $max_bandwidth_out;
            var $max_bandwidth_in_cfg, $max_bandwidth_out_cfg;
            var $targets = array();
            var $a_offset,             $b_offset;
            var $in_ds,                $out_ds;
            var $selected;
            var $inpercent,            $outpercent;
            var $inherit_fieldlist;
            var $vialist = array();

            var $outlinecolour;
            var $bwoutlinecolour;
            var $bwboxcolour;
            var $bwfontcolour;

            function WeatherMapLink() { $this->inherit_fieldlist=array
                (
                    'width' => 7,
                    'bwfont' => 2,
                    'arrowstyle' => 'classic',
                    'targets' => array(),
                    'infourl' => '',
                    'overliburl' => '',
                    'labelstyle' => 'percent',
                    'overlibwidth' => 0,
                    'overlibheight' => 0,
                    'max_bandwidth_in' => 100000000,
                    'max_bandwidth_out' => 100000000,
                    'outlinecolour' => array(0, 0, 0),
                    'bwoutlinecolour' => array(0, 0, 0),
                    'bwfontcolour' => array(0, 0, 0),
                    'bwboxcolour' => array(255, 255, 255),
                    'a_offset' => 'C',
                    'b_offset' => 'C',
                    'max_bandwidth_in_cfg' => '100M',
                    'max_bandwidth_out_cfg' => '100M'
                );
            // $this->a_offset = 'C';
            // $this->b_offset = 'C';
            //  $this->targets = array();
                }

            function Reset($newowner)
            {
                $this->owner=$newowner;

                if (isset($this->owner->defaultlink) && $this->name != 'DEFAULT') {
                // use the defaults from DEFAULT
                    $this->CopyFrom($this->owner->defaultlink); }
                else
                {
                    foreach (array_keys($this->inherit_fieldlist)as
                        $fld) { $this->$fld=$this->inherit_fieldlist[$fld]; }
                }
            }

            function CopyFrom($source)
            {
                foreach (array_keys($this->inherit_fieldlist)as $fld) { $this->$fld=$source->$fld; }
            }

            // Set the bandwidth for this link. Convert from KMGT as necessary
            function SetBandwidth($inbw, $outbw)
            {
                debug ("Setting bandwidth ($inbw, $outbw)\n");
                $this->max_bandwidth_in=unformat_number($inbw);
                $this->max_bandwidth_out=unformat_number($outbw);
                $this->max_bandwidth_in_cfg=$inbw;
                $this->max_bandwidth_out_cfg=$outbw;
            }

            function Draw($im, &$map)
            {
                #     $max_bw_in = $this->max_bandwidth_in;
                #     $max_bw_out = $this->max_bandwidth_out;

                $pc=(($this->bandwidth_out) / ($this->max_bandwidth_out)) * 100;
                $this->outpercent=$pc;

                $pc=(($this->bandwidth_in) / ($this->max_bandwidth_in)) * 100;
                $this->inpercent=$pc;

                $x1=$this->a->x;
                $y1=$this->a->y;

                $x2=$this->b->x;
                $y2=$this->b->y;

                //    print "$x1,$y1 -> $x2,$y2   via  ";

                $a_height=$map->nodes[$this->a->name]->height;
                $a_width=$map->nodes[$this->a->name]->width;

                $b_height=$map->nodes[$this->b->name]->height;
                $b_width=$map->nodes[$this->b->name]->width;

                //     print "[".$this->a_offset." ".$this->b_offset."] and ($a_width x $a_height and $b_width x $b_height)";

                list($dx, $dy)=calc_offset($this->a_offset, $a_width, $a_height);
                $x1+=$dx;
                $y1+=$dy;

                list($dx, $dy)=calc_offset($this->b_offset, $b_width, $b_height);
                $x2+=$dx;
                $y2+=$dy;

                //    print " becomes $x1,$y1 -> $x2,$y2";
                // print "\n";

                $outline_colour=NULL;

                if ($this->outlinecolour != array
                    (
                        -1,
                        -1,
                        -1
                    )) { $outline_colour=myimagecolorallocate(
                                             $im, $this->outlinecolour[0], $this->outlinecolour[1],
                                             $this->outlinecolour[2]); }

                $xpoints=array
                    (
                    );

                $ypoints=array
                    (
                    );

                $xpoints[]=$x1;
                $ypoints[]=$y1;

                // warn("There are VIAs.\n");
                foreach ($this->vialist as $via)
                {
                    // imagearc($im, $via[0],$via[1],20,20,0,360,$map->selected);
                    $xpoints[]=$via[0];
                    $ypoints[]=$via[1];
                }

                $xpoints[]=$x2;
                $ypoints[]=$y2;

                $link_in_colour=$map->ColourFromPercent($this->inpercent);
                $link_out_colour=$map->ColourFromPercent($this->outpercent);
                $link_width=$this->width;

                // for bulging animations
                if ($map->widthmod)
                    $link_width=($link_width * $task[4] * 1.5) / 100;

                list($q1_x,
                     $q1_y,
                     $q3_x,
                     $q3_y)=draw_curve($im,         $xpoints,        $ypoints,
                                       $link_width, $outline_colour, array($link_out_colour, $link_in_colour),
                                       $this->name, $map);

                if (!is_null($q1_x))
                {
                    $outbound=array
                        (
                            $q1_x,
                            $q1_y,
                            0,
                            0,
                            $this->outpercent,
                            $this->bandwidth_out
                        );

                    $inbound=array
                        (
                            $q3_x,
                            $q3_y,
                            0,
                            0,
                            $this->inpercent,
                            $this->bandwidth_in
                        );

                    if ($map->sizedebug)
                    {
                        $outbound[5]=$this->max_bandwidth_out;
                        $inbound[5]=$this->max_bandwidth_in;
                    }

                    foreach (array($inbound, $outbound)as $task)
                    {
                        $thelabel="";

                        if ($this->labelstyle != 'none')
                        {
                            if ($this->labelstyle == 'bits') { $thelabel=nice_bandwidth($task[5], $this->owner->kilo); }
                                elseif ($this->labelstyle == 'unformatted') { $thelabel=$task[5]; }
                                elseif ($this->labelstyle == 'percent') { $thelabel=format_number($task[4]) . "%"; }

                            $map->DrawLabel($im, $task[0],            $task[1],           $thelabel, $this->bwfont, 0,
                                            '',  $this->bwfontcolour, $this->bwboxcolour, $this->bwoutlinecolour);
                        }
                    }
                }
            }

            function WriteConfig($fd)
            {
                $output='';

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['infourl'] : $this->owner->defaultlink->infourl);

                if ($this->infourl != $comparison) { $output.="\tINFOURL " . $this->infourl . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['overliburl'] : $this->owner->defaultlink->overliburl);

                if ($this->overliburl != $comparison) { $output.="\tOVERLIBGRAPH " . $this->overliburl . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['overlibwidth'] : $this->owner->defaultlink->overlibwidth);

                if ($this->overlibwidth != $comparison) { $output.="\tOVERLIBWIDTH " . $this->overlibwidth . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['overlibheight'] : $this->owner->defaultlink->overlibheight);

                if ($this->overlibheight != $comparison) { $output.="\tOVERLIBHEIGHT " . $this->overlibheight . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['arrowstyle'] : $this->owner->defaultlink->arrowstyle);

                if ($this->arrowstyle != $comparison) { $output.="\tARROWSTYLE " . $this->arrowstyle . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['labelstyle'] : $this->owner->defaultlink->labelstyle);

                if ($this->labelstyle != $comparison) { $output.="\tBWLABEL " . $this->labelstyle . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['targets'] : $this->owner->defaultlink->targets);

                if ($this->targets != $comparison)
                {
                    $output.="\tTARGET";

                    foreach ($this->targets as $target) { $output.=" " . $target[4]; }

                    $output.="\n";
                }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['bwfont'] : $this->owner->defaultlink->bwfont);

                if ($this->bwfont != $comparison) { $output.="\tBWFONT " . $this->bwfont . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['width'] : $this->owner->defaultlink->width);

                if ($this->width != $comparison) { $output.="\tWIDTH " . $this->width . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['outlinecolour'] : $this->owner->defaultlink->outlinecolour);

                if ($this->outlinecolour != $comparison) { $output.="\tOUTLINECOLOR " . render_colour(
                                                                                            $this->outlinecolour)
                                                                        . "\n"; }

                $comparison=($this->name == 'DEFAULT' ? $this->inherit_fieldlist['bwoutlinecolour']
                                : $this->owner->defaultlink->bwoutlinecolour);

                if ($this->bwoutlinecolour != $comparison) { $output.="\tBWOUTLINECOLOR " . render_colour(
                                                                                                $this->bwoutlinecolour)
                                                                          . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['bwfontcolour'] : $this->owner->defaultlink->bwfontcolour);

                if ($this->bwfontcolour != $comparison) { $output.="\tBWFONTCOLOR " . render_colour(
                                                                                          $this->bwfontcolour) . "\n"; }

                $comparison=($this->name == 'DEFAULT'
                                ? $this->inherit_fieldlist['bwboxcolour'] : $this->owner->defaultlink->bwboxcolour);

                if ($this->bwboxcolour != $comparison) { $output.="\tBWBOXCOLOR " . render_colour(
                                                                                        $this->bwboxcolour) . "\n"; }

                if (isset($this->a) && isset($this->b))
                {
                    $output.="\tNODES " . $this->a->name;

                    if ($this->a_offset != 'C')
                        $output.=":" . $this->a_offset;

                    $output.=" " . $this->b->name;

                    if ($this->b_offset != 'C')
                        $output.=":" . $this->b_offset;

                    $output.="\n";
                }

                if (count($this->vialist) > 0)
                {
                    foreach ($this->vialist as $via)
                        $output.=sprintf("\tVIA %d %d\n", $via[0], $via[1]);
                }

                if (($this->max_bandwidth_in != $this->owner->defaultlink->max_bandwidth_in)
                    || ($this->max_bandwidth_out != $this->owner->defaultlink->max_bandwidth_out)
                    || ($this->name == 'DEFAULT'))
                {
                    if ($this->max_bandwidth_in == $this->max_bandwidth_out)
                        { $output.="\tBANDWIDTH " . $this->max_bandwidth_in_cfg . "\n"; }
                    else { $output
                        .="\tBANDWIDTH " . $this->max_bandwidth_in_cfg . " " . $this->max_bandwidth_out_cfg . "\n"; }
                }

                if ($output != '')
                {
                    fwrite($fd, "LINK " . $this->name . "\n");
                    fwrite($fd, "$output\n");
                }
            }

            function asJS()
            {
                $js='';
                $js.="Links[" . js_escape($this->name) . "] = {";

                if ($this->name != 'DEFAULT')
                {
                    $js.="a:'" . $this->a->name . "', ";
                    $js.="b:'" . $this->b->name . "', ";
                }

                $js.="width:'" . $this->width . "', ";
                $js.="target:";

                $tgt='';

                foreach ($this->targets as $target) { $tgt.=$target[4] . ' '; }

                $js.=js_escape(trim($tgt));
                $js.=",";

                $js.="bw_in:" . js_escape($this->max_bandwidth_in_cfg) . ", ";
                $js.="bw_out:" . js_escape($this->max_bandwidth_out_cfg) . ", ";

                $js.="name:" . js_escape($this->name) . ", ";
                $js.="overlibwidth:'" . $this->overlibheight . "', ";
                $js.="overlibheight:'" . $this->overlibwidth . "', ";
                $js.="infourl:" . js_escape($this->infourl) . ", ";
                $js.="overliburl:" . js_escape($this->overliburl);
                $js.="};\n";
                return $js;
            }
        
            function asJSON()
            {
                $js='';
                $js.="" . js_escape($this->name) . ": {";

                if ($this->name != 'DEFAULT')
                {
                    $js.="a:'" . $this->a->name . "', ";
                    $js.="b:'" . $this->b->name . "', ";
                }

                $js.="width:'" . $this->width . "', ";
                $js.="target:";

                $tgt='';

                foreach ($this->targets as $target) { $tgt.=$target[4] . ' '; }

                $js.=js_escape(trim($tgt));
                $js.=",";

                $js.="bw_in:" . js_escape($this->max_bandwidth_in_cfg) . ", ";
                $js.="bw_out:" . js_escape($this->max_bandwidth_out_cfg) . ", ";

                $js.="name:" . js_escape($this->name) . ", ";
                $js.="overlibwidth:'" . $this->overlibheight . "', ";
                $js.="overlibheight:'" . $this->overlibwidth . "', ";
                $js.="infourl:" . js_escape($this->infourl) . ", ";
                $js.="overliburl:" . js_escape($this->overliburl);
                $js.="},\n";
                return $js;
            }
        }

    ;

    class WeatherMap
        {
            var $nodes = array(); // an array of WeatherMapNodes
            var $links = array(); // an array of WeatherMapLinks
            var $texts = array(); // an array containing all the extraneous text bits
            var $used_images = array(); // an array of image filenames referred to (used by editor)
            var $background;
            var $htmlstyle;
            var $imap;
            var $colours;
            var $configfile;
            var $imagefile,
                $imageuri;
            var $rrdtool;
            var $title,
                $titlefont;
            var $kilo;
            var $sizedebug,
                $widthmod,
                $debugging;
            var $linkfont,
                $nodefont,
                $keyfont,
                $timefont;
            # var $bg_r, $bg_g, $bg_b;
            var $timex,
                $timey;
            var $width,
                $height;
            var $keyx,
                $keyy;
            var $titlex,
                $titley;
            var $keytext,
                $stamptext;
            var $htmloutputfile,
                $imageoutputfile;
            var $defaultlink,
                $defaultnode;
            var $need_size_precalc;
            var $keystyle;
            var $rrdtool_check;
            var $context;
            var $inherit_fieldlist;

            var $black,
                $white,
                $grey,
                $selected;

            function WeatherMap()
            {
                $this->inherit_fieldlist=array
                    (
                        'width' => 800,
                        'height' => 600,
                        'kilo' => 1000,
                        'numscales' => 0,
                        'dumpconfig' => FALSE,
                        'rrdtool_check' => '',
                        'context' => '',
                        'background' => '',
                        'imagefile' => '',
                        'htmlfile' => '',
                        'labelstyle' => 'percent', // redundant?
                        'htmlstyle' => 'static',
                        'keystyle' => 'classic',
                        'title' => 'Network Weathermap',
                        'keytext' => 'Traffic Load',
                        'stamptext' => 'Created: %b %d %Y %H:%M:%S',
                        'keyfont' => 4,
                        'titlefont' => 2,
                        'timefont' => 2,
                        'keyx' => -1,
                        'keyy' => -1,
                        'timex' => 0,
                        'timey' => 0,
                        'titlex' => -1,
                        'titley' => -1,
                        'sizedebug' => FALSE,
                        'debugging' => FALSE,
                        'widthmod' => FALSE
                    );

                $this->Reset();
            }

            function Reset()
            {
                foreach (array_keys($this->inherit_fieldlist)as $fld) { $this->$fld=$this->inherit_fieldlist[$fld]; }

                // these two are used for default settings
                $this->defaultlink=new WeatherMapLink;
                $this->defaultlink->name="DEFAULT";
                $this->defaultlink->Reset($this);

                $this->defaultnode=new WeatherMapNode;
                $this->defaultnode->name="DEFAULT";
                $this->defaultnode->Reset($this);

                $this->need_size_precalc=FALSE;

                $this->nodes=array
                    (
                    ); // an array of WeatherMapNodes

                $this->links=array
                    (
                    ); // an array of WeatherMapLinks

                $this->imap=new HTML_ImageMap('weathermap');
                $this->colours=array
                    (
                    );

                debug ("Adding default map colour set.\n");
                $defaults=array
                    (
                        'KEYTEXT' => array('bottom' => -2, 'top' => -1, 'red1' => 0, 'green1' => 0, 'blue1' => 0),
                        'KEYBG' => array('bottom' => -2, 'top' => -1, 'red1' => 255, 'green1' => 255, 'blue1' => 255),
                        'BG' => array('bottom' => -2, 'top' => -1, 'red1' => 255, 'green1' => 255, 'blue1' => 255),
                        'TITLE' => array('bottom' => -2, 'top' => -1, 'red1' => 0, 'green1' => 0, 'blue1' => 0),
                        'TIME' => array('bottom' => -2, 'top' => -1, 'red1' => 0, 'green1' => 0, 'blue1' => 0)
                    );

                foreach ($defaults as $key => $def) { $this->colours[$key]=$def; }

                $this->configfile='';
                $this->imagefile='';
                $this->imageuri='';

                # $this->bg_r = 255;
                # $this->bg_g = 255;
                # $this->bg_b = 255;

                $this->fonts=array
                    (
                    );
            }

            function myimagestring($image, $fontnumber, $x, $y, $string, $colour)
            {
                // if it's supposed to be a special font, and it hasn't been defined, then fall through
                if ($fontnumber > 5 && !isset($this->fonts[$fontnumber]))
                {
                    warn ("Using a non-existent special font ($fontnumber) - falling back to internal GD fonts\n");

                    $fontnumber=5;
                }

                if (($fontnumber > 0) && ($fontnumber < 6))
                    { imagestring($image, $fontnumber, $x, $y - imagefontheight($fontnumber), $string, $colour); }
                else
                {
                    // look up what font is defined for this slot number
                    if ($this->fonts[$fontnumber]->type == 'truetype')
                        { imagettftext($image, $this->fonts[$fontnumber]->size, 0, $x, $y,
                                       $colour, $this->fonts[$fontnumber]->file, $string); }

                    if ($this->fonts[$fontnumber]->type == 'gd')
                        { imagestring($image, $this->fonts[$fontnumber]->gdnumber,
                                      $x,      $y - imagefontheight($this->fonts[$fontnumber]->gdnumber),
                                      $string, $colour); }
                }
            }

            function myimagestringsize($fontnumber, $string)
            {
                if (($fontnumber > 0) && ($fontnumber < 6))
                    { return array(imagefontwidth($fontnumber) * strlen($string), imagefontheight($fontnumber)); }
                else
                {
                    // look up what font is defined for this slot number
                    if (!isset($this->fonts[$fontnumber]))
                    {
                        warn ("Using a non-existent special font ($fontnumber) - falling back to internal GD fonts\n");
                        $fontnumber=5;
                        return array(imagefontwidth($fontnumber) * strlen($string), imagefontheight($fontnumber));
                    }
                    else
                    {
                        if ($this->fonts[$fontnumber]->type == 'truetype')
                        {
                            $bounds=imagettfbbox($this->fonts[$fontnumber]->size, 0, $this->fonts[$fontnumber]->file,
                                                 $string);
                            return (array($bounds[4] - $bounds[0], $bounds[1] - $bounds[5]));
                        }

                        if ($this->fonts[$fontnumber]->type == 'gd')
                            { return array(imagefontwidth($this->fonts[$fontnumber]->gdnumber) * strlen($string),
                                         imagefontheight($this->fonts[$fontnumber]->gdnumber)); }
                    }
                }
            }

            function RandomData()
            {
                foreach ($this->links as $link)
                {
                    $this->links[$link->name]->bandwidth_in=rand(0, $link->max_bandwidth_in);
                    $this->links[$link->name]->bandwidth_out=rand(0, $link->max_bandwidth_out);
                }
            }

            function ReadFromTabbedText($target, $label)
            {
                debug ("ReadFromTabbedText: Opening $target \n");

                $inbw=-1;
                $outbw=-1;
                $matches=0;

                $fd=fopen($target, "r");

                if ($fd)
                {
                    while (!feof($fd))
                    {
                        $buffer=fgets($fd, 4096);
                        # strip out any Windows line-endings that have gotten in here
                        $buffer=str_replace("\r", "", $buffer);

                        if (preg_match("/^$label\t(\d+\.?\d*[KMGT]*)\t(\d+\.?\d*[KMGT]*)/", $buffer, $matches))
                        {
                            $inbw=unformat_number($matches[1]);
                            $outbw=unformat_number($matches[2]);
                        }
                    }
                }
                else {
                // some error code to go in here
                    debug ("ReadFromTabbedText: Couldn't open ($target). \n"); }

                debug ("ReadFromTabbedText: Returning ($inbw,$outbw)\n");
                return (array($inbw, $outbw));
            }

            function ReadFromMRTG($target)
            {
                debug ("ReadFromMRTG: Opening $target \n");

                $inbw=-1;
                $outbw=-1;
                $matches=0;

                $fd=fopen($target, "r");

                if ($fd)
                {
                    while (!feof($fd))
                    {
                        $buffer=fgets($fd, 4096);

                        if (preg_match("/<\!-- cuin d (\d+) -->/", $buffer, $matches)) { $inbw=$matches[1] * 8; }

                        if (preg_match("/<\!-- cuout d (\d+) -->/", $buffer, $matches)) { $outbw=$matches[1] * 8; }
                    }
                }
                else {
                // some error code to go in here
                    debug ("ReadFromMRTG: Couldn't open ($target). \n"); }

                debug ("ReadFromMRTG: Returning ($inbw,$outbw)\n");
                return (array($inbw, $outbw));
            }

            function ReadFromRRD($target, $in_ds = "traffic_in", $out_ds = "traffic_out")
            {
            		
            	$inbw = -1;
                $outbw = -1;

                // rrdtool fetch RRDfile.rrd AVERAGE --start now
                // we get the last 400 seconds of data - this might be 1 or 2 lines, depending on when in the
                // cacti polling cycle we get run. This ought to stop the 'some lines are grey' problem that some
                // people were seeing

                if(file_exists($target)){
                    if ($this->rrdtool_check == 'FOUND'){
                        debug("ReadFromRRD: Target DS names are $in_ds and $out_ds\n");
                        $command = $this->rrdtool . " fetch $target AVERAGE --start now-800";
                        debug("ReadFromRRD: Running: $command\n");
                        $pipe = popen($command, "r");
                        $lines = array();
                        $count = 0;
                        if (isset($pipe))
                        {
                            $headings=fgets($pipe, 4096);
                            fgets($pipe, 4096); // skip the blank line
                            $buffer='';
    
                            while (!feof($pipe))   {
                                $line=fgets($pipe, 4096);
                                debug ("> " . $line);
                                $buffer.=$line;
                                $lines[]=$line;
                            }
                            
                            $rlines=array_reverse($lines);
                            $gotline=0;
                            $theline='';
    
                            foreach ($rlines as $line)
                            {
                                debug ("--" . $line . "\n");
                                $cols=preg_split("/\s+/", $line);
                                $dataok=1;
                                $ii=0;
    
                                foreach ($cols as $col)
                                {
                                    #                       if( ! is_numeric($col) ) { $dataok=0; }
                                    if (trim($col) != '' && !preg_match('/^\d+\.?\d*e?[+-]?\d*:?$/i', $col))
                                    {
                                        $dataok=0;
                                        debug ("ReadFromRRD: $ii: This isn't a number: [$col]\n");
                                    }
    
                                    # if($col=='nan') { $dataok=0; }
                                    $ii++;
                                }
    
                                if ($gotline == 0 && $dataok == 1 && trim($line) != '')
                                {
                                    debug ("ReadFromRRD: Found a good line: $line ($headings)\n");
                                    $theline=$line;
                                    $gotline=1;
                                    $countwas=$count;
                                }
    
                                $count++;
                            }
    
                            pclose ($pipe);
                        }
    
                        if ($theline != ''){
                            if ($countwas > 2) { 
                            	warn("ReadFromRRD: Data is not most recent entry ($countwas) for link: $target\n"); 
                            }
                            debug ("ReadFromRRD: Our line is $theline\n");
                            $cols=preg_split("/\s+/", $theline);
                            $heads=preg_split("/\s+/", $headings);
    
                            # $values = array_combine($heads,$cols);
                            for ($i=0, $cnt=count($cols); $i < $cnt; $i++) { 	
                            	$values[$heads[$i]]=$cols[$i]; 
                            }
    
                            // as long as no-one actually manages to create an RRD with a DS of '-', then this will just fall through to 0 for '-'
                            if (isset($values[$in_ds]))
 	                           	$inbw = $values[$in_ds];
    						if (isset($values[$out_ds]))
 	                           $outbw = $values[$out_ds];
                        }
                    } else { 	
                 		warn("rrdtool path is incorrect. Check 'weathermap' script, line 27.\n"); 
                 	}
                } else {
                	warn("Target $target doesn't exist. Is it a file?\n");
                }
                debug ("ReadFromRRD: Returning ($inbw,$outbw)\n");
                return (array($inbw, $outbw));
            }

            function ReadData()
            {
                debug("================== ReadData: Updating link data for all links\n");

                if ($this->sizedebug == 0)
                {
                    foreach ($this->links as $link)
                    {
                        $total_in = 0;
                        $total_out = 0;
                        $name = $link->name;
                        debug("\n\nReadData for Link $name: \n");
                        if (!is_null(($link->targets))){                        	
                        	foreach ($link->targets as $target){
                                debug ("  New Target: $target[4] ($target[0])\n");
                                if ($target[4] != '')
                                {
                                    if (preg_match("/\.rrd$/i", $target[0])){ // Direct from RRD
                                        debug("ReadData: Updating link $name from RRD Target $target[4]\n");
                                        list($in, $out)=$this->ReadFromRRD($target[0], $target[1], $target[2]);
                                        list($in, $out)=$this->ReadFromRRD($target[0], $target[1], $target[2]);
                                    } 
                                    if (isset($in) && $in != -1)
                                    	$total_in = $total_in + $in;
                                    if (isset($out) && $out != -1)
	                                    $total_out = $total_out + $out;
                                }
                            }
                        }

                        $this->links[$name]->bandwidth_in = $total_in;
                        $this->links[$name]->bandwidth_out = $total_out;
                        debug ("ReadData: Setting $total_in,$total_out\n");
                    }
                    debug ("\nReadData Completed.\n--------------\n");
                }
            }

            // nodename is a vestigal parameter, from the days when nodes where just big labels
            function DrawLabel($im, $x, $y, $text, $font, $padding, $nodename, $textcolour, $bgcolour, $outlinecolour)
            {
                list($strwidth, $strheight)=$this->myimagestringsize($font, $text);

                $extra=3;

                $x1=$x - ($strwidth / 2) - $padding - $extra;
                $x2=$x + ($strwidth / 2) + $padding + $extra;
                $y1=$y - ($strheight / 2) - $padding - $extra;
                $y2=$y + ($strheight / 2) + $padding + $extra;

                if ($bgcolour != array
                    (
                        -1,
                        -1,
                        -1
                    ))
                {
                    $bgcol=myimagecolorallocate($im, $bgcolour[0], $bgcolour[1], $bgcolour[2]);
                    imagefilledrectangle($im, $x1, $y1, $x2, $y2, $bgcol);
                }

                if ($outlinecolour != array
                    (
                        -1,
                        -1,
                        -1
                    ))
                {
                    $outlinecol=myimagecolorallocate($im, $outlinecolour[0], $outlinecolour[1], $outlinecolour[2]);
                    imagerectangle($im, $x1, $y1, $x2, $y2, $outlinecol);
                }

                $textcol=myimagecolorallocate($im, $textcolour[0], $textcolour[1], $textcolour[2]);
                $this->myimagestring($im, $font, $x - $strwidth / 2, $y + $strheight / 2 + 1, $text, $textcol);
            }

            function ColourFromPercent($percent)
            {
                $colours=$this->colours;

                if ($percent > 100)
                {
                    warn ("Clipped $percent% to 100%\n");
                    $percent=100;
                }

                foreach ($colours as $key => $colour)
                {
                    if (($percent >= $colour['bottom']) and ($percent <= $colour['top']))
                    {
                        if (isset($colour['red2']))
                        {
                            $ratio=($percent - $colour["bottom"]) / ($colour["top"] - $colour["bottom"]);
                            // print "GRADIENT SCALE ($ratio between ".$colour["bottom"]." and ".$colour["top"].")\n";

                            $r=$colour["red1"] + ($colour["red2"] - $colour["red1"]) * $ratio;
                            $g=$colour["green1"] + ($colour["green2"] - $colour["green1"]) * $ratio;
                            $b=$colour["blue1"] + ($colour["blue2"] - $colour["blue1"]) * $ratio;

                            // print "    returning $r,$g,$b\n";
                            return myimagecolorallocate($this->image, $r, $g, $b);
                        }
                        else { return $colour['gdref1']; }
                    }
                }

                // you'll only get grey for a COMPLETELY quiet link if there's no 0 in the SCALE lines
                if ($percent == 0) { return $this->grey; }

                // and you'll only get white for a link with no colour assigned
                return $this->white;
            }

            function coloursort($a, $b)
            {
                if ($a['bottom'] == $b['bottom']) { return 0; }

                if ($a['bottom'] < $b['bottom']) { return -1; }

                return 1;
            }

            function DrawLegend_Classic($im)
            {
                $title=$this->keytext;

                $colours=$this->colours;
                $nscales=$this->numscales;

                $font=$this->keyfont;

                $x=$this->keyx;
                $y=$this->keyy;

                list($tilewidth, $tileheight)=$this->myimagestringsize($font, "MMMM");
                $tileheight=$tileheight * 1.1;
                $tilespacing=$tileheight + 2;

                if (($x >= 0) && ($y >= 0))
                {
                    //    $minwidth = imagefontwidth($font) * strlen('XX 100%-100%')+10;
                    //    $boxwidth = imagefontwidth($font) * strlen($title) + 10;
                    list($minwidth, $junk)=$this->myimagestringsize($font, 'MMMM 100%-100%');
                    list($boxwidth, $junk)=$this->myimagestringsize($font, $title);

                    $minwidth+=10;
                    $boxwidth+=10;

                    if ($boxwidth < $minwidth) { $boxwidth=$minwidth; }

                    $boxheight=$tilespacing * ($nscales + 1) + 10;

                    $boxx=$x;
                    $boxy=$y;

                    # allow for X11-style negative positioning
                    if ($boxx < 0) { $boxx+=$this->width; }

                    if ($boxy < 0) { $boxy+=$this->height; }

                    imagefilledrectangle($im, $boxx, $boxy, $boxx + $boxwidth, $boxy + $boxheight,
                                         $this->colours['KEYBG']['gdref1']);
                    imagerectangle($im, $boxx, $boxy, $boxx + $boxwidth, $boxy + $boxheight,
                                   $this->colours['KEYTEXT']['gdref1']);
                    $this->myimagestring($im, $font, $boxx + 4, $boxy + 4 + $tileheight, $title,
                                         $this->colours['KEYTEXT']['gdref1']);

                    usort($colours, array("Weathermap", "coloursort"));

                    $i=1;

                    foreach ($colours as $colour)
                    {
                        if ($colour['bottom'] >= 0)
                        {
                            $y=$boxy + $tilespacing * $i + 8;
                            $x=$boxx + 6;

                            if (isset($colour['red2']))
                            {
                                for ($n=0; $n < $tilewidth; $n++)
                                {
                                    $percent
                                        =$colour['bottom'] + ($n / $tilewidth) * ($colour['top'] - $colour['bottom']);
                                    imagefilledrectangle($im, $x + $n, $y, $x + $n, $y + $tileheight,
                                                         $this->ColourFromPercent($percent));
                                }
                            }
                            else
                            {
                                # pick a percentage in the middle...
                                $percent=($colour['bottom'] + $colour['top']) / 2;
                                imagefilledrectangle($im, $x, $y, $x + $tilewidth, $y + $tileheight,
                                                     $this->ColourFromPercent($percent));
                            }

                            $labelstring=sprintf("%s-%s%%", $colour['bottom'], $colour['top']);
                            $this->myimagestring($im, $font, $x + 4 + $tilewidth, $y + $tileheight, $labelstring,
                                                 $this->colours['KEYTEXT']['gdref1']);
                            $i++;
                        }
                    }

                    $this->imap->addArea("Rectangle", "LEGEND", '',
                                         array($boxx, $boxy, $boxx + $boxwidth, $boxy + $boxheight));
                //   $this->imap->setProp("href","#","LEGEND");
                //   $this->imap->setProp("extrahtml","onclick=\"position_legend();\"","LEGEND");

                }
            }

            function DrawTimestamp($im, $font, $colour)
            {
                # add a timestamp to the corner, so we can tell if it's all being updated
                # $datestring = "Created: ".date("M d Y H:i:s",time());
                $datestring=strftime($this->stamptext, time());

                list($boxwidth, $boxheight)=$this->myimagestringsize($font, $datestring);

                $x=$this->width - $boxwidth;
                $y=$boxheight;
                ;

                if (($this->timex > 0) && ($this->timey > 0))
                {
                    $x=$this->timex;
                    $y=$this->timey;
                }

                $this->myimagestring($im, $font, $x, $y, $datestring, $colour);

                $this->imap->addArea("Rectangle", "TIMESTAMP", '', array($x, $y, $x + $boxwidth, $y - $boxheight));
            }

            function DrawTitle($im, $font, $colour)
            {
                $string=$this->title;

                list($boxwidth, $boxheight)=$this->myimagestringsize($font, $string);

                $x=10;
                $y=$this->titley - $boxheight;

                if (($this->titlex >= 0) && ($this->titley >= 0))
                {
                    $x=$this->titlex;
                    $y=$this->titley;
                }

                $this->myimagestring($im, $font, $x, $y, $string, $colour);

                $this->imap->addArea("Rectangle", "TITLE", '', array($x, $y, $x + $boxwidth, $y - $boxheight));
            }

            function ReadConfig($filename)
            {
                $curnode=null;
                $curlink=null;
                $matches=0;
                $nodesseen=0;
                $linksseen=0;
                $scalesseen=0;

                $last_seen="---";

//echo getcwd().$filename."<br>";

                $fd=fopen($filename, "r");

                if ($fd)
                {
                    $linecount=0;

                    while (!feof($fd))
                    {
                        $buffer=fgets($fd, 4096);
                        # strip out any Windows line-endings that have gotten in here
                        $buffer=str_replace("\r", "", $buffer);
                        $linematched=0;
                        $linecount++;

                        if (preg_match("/^\s*#/", $buffer)) {
                        // this is a comment line
                            }
                        else
                        {
                            if (preg_match("/^\s*(LINK|NODE)\s+(\S+)/i", $buffer, $matches))
                            {
                                // first, save the previous item, before starting work on the new one
                                if ($last_seen == "NODE")
                                {
                                    if ($curnode->name == 'DEFAULT')
                                    {
                                        $this->defaultnode=$curnode;
                                        debug ("Saving Default Node: " . $curnode->name . "\n");
                                    }
                                    else
                                    {
                                        $this->nodes[$curnode->name]=$curnode;
                                        debug ("Saving Node: " . $curnode->name . "\n");
                                    }
                                }

                                if ($last_seen == "LINK")
                                {
                                    if ($curlink->name == 'DEFAULT')
                                    {
                                        $this->defaultlink=$curlink;
                                        debug ("Saving Default Link: " . $curlink->name . "\n");
                                    }
                                    else
                                    {
                                        if (isset($curlink->a) && isset($curlink->b))
                                        {
                                            $this->links[$curlink->name]=$curlink;
                                            debug ("Saving Link: " . $curlink->name . "\n");
                                        }
                                        else { warn
                                            ("Dropping LINK " . $curlink->name . " - it hasn't got 2 NODES!\n"); }
                                    }
                                }

                                if ($matches[1] == 'LINK')
                                {
                                    if ($matches[2] == 'DEFAULT')
                                    {
                                        if ($linksseen > 0) { warn
                                            ("LINK DEFAULT is not the first LINK. Defaults will not apply to earlier LINKs.\n");
                                            }

                                        $curlink=$this->defaultlink;
                                    }
                                    else
                                    {
                                        $curlink=new WeatherMapLink;
                                        $curlink->name=$matches[2];
                                        $curlink->Reset($this);
                                        $linksseen++;
                                    }

                                    $last_seen="LINK";
                                    $linematched++;
                                }

                                if ($matches[1] == 'NODE')
                                {
                                    if ($matches[2] == 'DEFAULT')
                                    {
                                        if ($nodesseen > 0) { warn
                                            ("NODE DEFAULT is not the first NODE. Defaults will not apply to earlier NODEs.\n");
                                            }

                                        $curnode=$this->defaultnode;
                                    }
                                    else
                                    {
                                        $curnode=new WeatherMapNode;
                                        $curnode->name=$matches[2];
                                        $curnode->Reset($this);
                                        $nodesseen++;
                                    }

                                    $last_seen="NODE";
                                    $linematched++;
                                }
                            }

                            if (preg_match("/^\s*POSITION\s+(\d+)\s+(\d+)/i", $buffer, $matches))
                            {
                                if ($last_seen == 'NODE')
                                {
                                    $curnode->x=$matches[1];
                                    $curnode->y=$matches[2];
                                    $linematched++;
                                }
                            }

                            if (preg_match("/^\s*LABEL\s+(.*)/i", $buffer, $matches))
                            {
                                if ($last_seen == 'NODE')
                                {
                                    $curnode->label=$matches[1];
                                    $linematched++;
                                }
                            }

                            if (preg_match("/^\s*NODES\s+(\S+)\s+(\S+)/i", $buffer, $matches))
                            {
                                if ($last_seen == 'LINK')
                                {
                                    $valid_nodes=2;

                                    foreach (array(1, 2)as $i)
                                    {
                                        $endoffset[$i]='C';
                                        $nodenames[$i]=$matches[$i];

                                        if (preg_match("/:(NE|SE|NW|SW|N|S|E|W)$/i", $matches[$i], $submatches))
                                        {
                                            $endoffset[$i]=$submatches[1];
                                            $nodenames[$i]=preg_replace("/:(NE|SE|NW|SW|N|S|E|W)$/i", '', $matches[$i]);
                                            $this->need_size_precalc=TRUE;
                                        }

                                        if (!array_key_exists($nodenames[$i], $this->nodes))
                                        {
                                            warn ("Unknown node '" . $nodenames[$i]
                                                     . "' on line $linecount of config\n");
                                            $valid_nodes--;
                                        }
                                    }

                                    // TODO - really, this should kill the whole link, and reset for the next one
                                    if ($valid_nodes == 2)
                                    {
                                        $curlink->a=$this->nodes[$nodenames[1]];
                                        $curlink->b=$this->nodes[$nodenames[2]];
                                        $curlink->a_offset=$endoffset[1];
                                        $curlink->b_offset=$endoffset[2];
                                    }
                                    else {
                                    // this'll stop the current link being added
                                        $last_seen="broken"; }

                                    $linematched++;
                                }
                            }

                            if (preg_match("/^\s*TARGET\s+(.*)/i", $buffer, $matches))
                            {
                                if ($last_seen == 'LINK')
                                {
                                    $linematched++;

                                    $targets=preg_split('/\s+/', $matches[1], -1, PREG_SPLIT_NO_EMPTY);

                                    foreach ($targets as $target)
                                    {

                                        // we store the original TARGET string, and line number, along with the breakdown, to make nicer error messages later
                                        $newtarget=array
                                            (
                                                $target,
                                                'traffic_in',
                                                'traffic_out',
                                                $linecount,
                                                $target
                                            );

                                        //   $curlink->target = $target;

                                        # if it's an RRD file, then allow for the user to specify the
                                        # DSs to be used. The default is traffic_in, traffic_out, which is
                                        # OK for Cacti (most of the time), but if you have other RRDs...
                                        if (preg_match("/(.*\.rrd):([\-a-zA-Z0-9_]+):([\-a-zA-Z0-9_]+)$/i", $target,
                                                       $matches))
                                        {
                                            $newtarget[0]=$matches[1];
                                            $newtarget[1]=$matches[2];
                                            $newtarget[2]=$matches[3];
                                        }

                                        if (preg_match("/\.rrd$/i", $newtarget[0]) && ($this->context != 'editor') && ($this->rrdtool_check == ''))
                                        {
                                            // if this is the first .rrd TARGET, then we better check we have rrdtool
                                            // (delayed to here, so that MRTG-only people don't have to have it too)
                                            debug ("RRDTOOL check triggered\n");

                                            if (file_exists($this->rrdtool)) { $this->rrdtool_check='FOUND'; }
                                            else
                                            {
                                                warn
                                                    ("ERROR: Can't find RRDTOOL. Check line 27 of the 'weathermap' script.\nRRD-based TARGETs will fail.\n");
                                                $this->rrdtool_check='FAILED';
                                            }
                                        }

                                        // now we've (maybe) messed with it, we'll store the array of target specs
                                        $curlink->targets[]=$newtarget;
                                    }
                                }
                            }

                            if (preg_match("/^\s*WIDTH\s+(\d+)/i", $buffer, $matches))
                            {
                                if ($last_seen == 'LINK')
                                {
                                    $curlink->width=$matches[1];
                                    $linematched++;
                                }
                                else
                                {
                                    $this->width=$matches[1];
                                    $linematched++;
                                }
                            }

                            if (preg_match("/^\s*HEIGHT\s+(\d+)/i", $buffer, $matches))
                            {
                                $this->height=$matches[1];
                                $linematched++;
                            }

                            if (preg_match("/^\s*BANDWIDTH\s+(\d+\.?\d*[KMGT]*)\s*$/i", $buffer, $matches))
                            {
                                if ($last_seen == 'LINK')
                                {
                                    $curlink->SetBandwidth($matches[1], $matches[1]);
                                    $linematched++;
                                }
                            }

                            if (preg_match("/^\s*BANDWIDTH\s+(\d+\.?\d*[KMGT]*)\s+(\d+\.?\d*[KMGT]*)\s*$/i", $buffer,
                                           $matches))
                            {
                                if ($last_seen == 'LINK')
                                {
                                    $curlink->SetBandwidth($matches[1], $matches[2]);
                                    $linematched++;
                                }
                            }

                            if (preg_match("/^\s*ICON\s+(\S+)\s*$/i", $buffer, $matches))
                            {
                                if ($last_seen == 'NODE')
                                {
                                    $curnode->iconfile=$matches[1];
                                    $this->used_images[] = $matches[1];
                                    $linematched++;
                                }
                            }

                            if (preg_match("/^\s*INFOURL\s+(\S+)/i", $buffer, $matches))
                            {
                                if ($last_seen == "LINK") { $curlink->infourl=$matches[1]; }
                                else { $curnode->infourl=$matches[1]; }

                                $linematched++;
                            }

                            if (preg_match("/^\s*OVERLIBGRAPH\s+(\S+)/i", $buffer, $matches))
                            {
                                if ($last_seen == "LINK") { $curlink->overliburl=$matches[1]; }
                                else { $curnode->overliburl=$matches[1]; }

                                $linematched++;
                            }

                            if (preg_match("/^\s*OVERLIBHEIGHT\s+(\d+)/i", $buffer, $matches))
                            {
                                if ($last_seen == "LINK") { $curlink->overlibheight=$matches[1]; }
                                else { $curnode->overlibheight=$matches[1]; }

                                $linematched++;
                            }

                            if (preg_match("/^\s*OVERLIBWIDTH\s+(\d+)/i", $buffer, $matches))
                            {
                                if ($last_seen == "LINK") { $curlink->overlibwidth=$matches[1]; }
                                else { $curnode->overlibwidth=$matches[1]; }

                                $linematched++;
                            }

                            if ($last_seen == 'NODE' && preg_match("/^\s*LABELFONT\s+(\d+)/i", $buffer, $matches))
                            {
                                $curnode->labelfont=$matches[1];
                                $linematched++;
                            }

                            if ($last_seen == 'NODE' && preg_match(
                                                            "/^\s*LABELOFFSET\s+([-+]*\d+)\s+([-+]*\d+)/i", $buffer,
                                                            $matches))
                            {
                                $curnode->labeloffsetx=$matches[1];
                                $curnode->labeloffsety=$matches[2];

                                $linematched++;
                            }

                            if ($last_seen == 'NODE' && preg_match(
                                                            "/^\s*LABELOFFSET\s+(NE|SE|NW|SW|N|S|E|W)/i", $buffer,
                                                            $matches))
                            {
                                $curnode->labeloffset=$matches[1];
                                $linematched++;
                            }

                            if ($last_seen == 'LINK' && preg_match("/^\s*VIA\s+(\d+)\s+(\d+)/i", $buffer, $matches))
                            {
                                $curlink->vialist[]=array
                                    (
                                        $matches[1],
                                        $matches[2]
                                    );

                                $linematched++;
                            }

                            if ($last_seen == 'LINK' && preg_match("/^\s*BWFONT\s+(\d+)/i", $buffer, $matches))
                            {
                                $curlink->bwfont=$matches[1];
                                $linematched++;
                            }

                            if ($last_seen == 'LINK' && preg_match(
                                                            "/^\s*BWLABEL\s+(bits|percent|unformatted|none)/i", $buffer,
                                                            $matches))
                            {
                                $curlink->labelstyle=strtolower($matches[1]);
                                $linematched++;
                            }

                            if (preg_match(
                                    "/^\s*SCALE\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/i",
                                    $buffer,
                                    $matches))
                            {
                                $key=$matches[1] . ':' . $matches[2];
                                $this->colours[$key]['bottom']=$matches[1];
                                $this->colours[$key]['top']=$matches[2];
                                $this->colours[$key]['red1']=$matches[3];
                                $this->colours[$key]['green1']=$matches[4];
                                $this->colours[$key]['blue1']=$matches[5];
                                $this->colours[$key]['red2']=$matches[6];
                                $this->colours[$key]['green2']=$matches[7];
                                $this->colours[$key]['blue2']=$matches[8];
                                $this->colours[$key]['key']=$key;
                                $scalesseen++;
                                $linematched++;
                            }

                            if (preg_match("/^\s*SCALE\s+(\d+\.?\d*)\s+(\d+\.?\d*)\s+(\d+)\s+(\d+)\s+(\d+)/i", $buffer,
                                           $matches) && ($linematched == 0))
                            {
                                $key=$matches[1] . ':' . $matches[2];
                                $this->colours[$key]['bottom']=$matches[1];
                                $this->colours[$key]['top']=$matches[2];
                                $this->colours[$key]['red1']=$matches[3];
                                $this->colours[$key]['green1']=$matches[4];
                                $this->colours[$key]['blue1']=$matches[5];
                                $this->colours[$key]['key']=$key;
                                $scalesseen++;
                                $linematched++;
                            }

                            if (preg_match("/^\s*KEYPOS\s+(\d+)\s+(\d+)(.*)/i", $buffer, $matches))
                            {
                                $this->keyx=$matches[1];
                                $this->keyy=$matches[2];
                                $extra=trim($matches[3]);

                                if ($extra != '')
                                    $this->keytext=$extra;

                                $linematched++;
                            }

                            if (preg_match("/^\s*TITLEPOS\s+(\d+)\s+(\d+)\s+(.*)/i", $buffer, $matches))
                            {
                                $this->titlex=$matches[1];
                                $this->titley=$matches[2];
                                $extra=trim($matches[3]);

                                if ($extra != '')
                                    $this->title=$extra;

                                $linematched++;
                            }

                            # truetype font definition (actually, we don't really check if it's truetype) - filename + size
                            if (preg_match("/^\s*FONTDEFINE\s+(\d+)\s+(\S+)\s+(\d+)\s*$/i", $buffer, $matches))
                            {
                                if (function_exists("imagettfbbox"))
                                {
                                    // test if this font is valid, before adding it to the font table...
                                    $bounds=@imagettfbbox($matches[3], 0, $matches[2], "Ignore me");

                                    if (isset($bounds[0]))
                                    {
                                        $this->fonts[$matches[1]]->type="truetype";
                                        $this->fonts[$matches[1]]->file=$matches[2];
                                        $this->fonts[$matches[1]]->size=$matches[3];
                                    }
                                    else { warn
                                        ("Failed to load ttf font " . $matches[2] . " - at config line $linecount\n"); }
                                }
                                else { warn
                                    ("imagettfbbox() is not a defined function. You don't seem to have FreeType compiled into your gd module.\n");
                                    }

                                $linematched++;
                            }

                            # GD font definition (no size here)
                            if (preg_match("/^\s*FONTDEFINE\s+(\d+)\s+(\S+)\s*$/i", $buffer, $matches))
                            {
                                $newfont=imageloadfont($matches[2]);

                                if ($newfont)
                                {
                                    $this->fonts[$matches[1]]->type="gd";
                                    $this->fonts[$matches[1]]->file=$matches[2];
                                    $this->fonts[$matches[1]]->gdnumber=$newfont;
                                }
                                else { warn ("Failed to load GD font: " . $matches[2]
                                                . " ($newfont) at config line $linecount\n"); }

                                $linematched++;
                            }

                            if (preg_match("/^\s*KEYFONT\s+(\d+)/i", $buffer, $matches))
                            {
                                $this->keyfont=$matches[1];
                                $linematched++;
                            }

                            if (preg_match("/^\s*TIMEFONT\s+(\d+)/i", $buffer, $matches))
                            {
                                $this->timefont=$matches[1];
                                $linematched++;
                            }

                            if (preg_match("/^\s*TITLEFONT\s+(\d+)/i", $buffer, $matches))
                            {
                                $this->titlefont=$matches[1];
                                $linematched++;
                            }

                            if (preg_match("/^\s*NODEFONT\s+(\d+)/i", $buffer, $matches))
                            {
                                $this->nodefont=$matches[1];
                                $this->defaultnode->labelfont=$matches[1];
                                warn
                                    ("NODEFONT is deprecated. Use NODE DEFAULT and LABELFONT instead. config line $linecount\n");
                                $linematched++;
                            }

                            if (preg_match("/^\s*LINKFONT\s+(\d+)/i", $buffer, $matches))
                            {
                                $this->linkfont=$matches[1];
                                $this->defaultlink->bwfont=$matches[1];
                                warn
                                    ("LINKFONT is deprecated. Use LINK DEFAULT and BWFONT instead. config line $linecount\n");
                                $linematched++;
                            }

                            if (preg_match("/^\s*TIMEPOS\s+(\d+)\s+(\d+)(.*)/i", $buffer, $matches))
                            {
                                $this->timex=$matches[1];
                                $this->timey=$matches[2];
                                $extra=trim($matches[3]);

                                if ($extra != '')
                                    $this->stamptext=$extra;

                                $linematched++;
                            }

                            //if(preg_match("/^\s*KEYSTYLE\s+(classic|horizontal|vertical)/i",$buffer, $matches))
                            //{
                            //  $this->keystyle = strtolower($matches[1]);
                            //  $linematched++;
                            //}

                            if (preg_match("/^\s*BWLABELS\s+(bits|percent|none)/i", $buffer, $matches))
                            {
                                # $this->labelstyle = strtolower($matches[1]);
                                $this->defaultlink->labelstyle=strtolower($matches[1]);
                                warn
                                    ("BWLABELS is deprecated. Use LINK DEFAULT and BWLABEL instead. config line $linecount\n");

                                $linematched++;
                            }

                            if (preg_match("/^\s*KILO\s+(\d+)/i", $buffer, $matches))
                            {
                                $this->kilo=$matches[1];
                                $linematched++;
                            }

                            if (preg_match("/^\s*BACKGROUND\s+(\S+)/i", $buffer, $matches))
                            {
                                $this->background=$matches[1];
                                $this->used_images[] = $matches[1];
                                $linematched++;
                            }

                            if (preg_match(
                                    "/^\s*(TIME|TITLE|KEYBG|KEYTEXT|KEYOUTLINE|BG)COLOR\s+(\d+)\s+(\d+)\s+(\d+)/i",
                                    $buffer,
                                    $matches))
                            {
                                $key=$matches[1];
                                # "Found colour line for $key\n";
                                $this->colours[$key]['red1']=$matches[2];
                                $this->colours[$key]['green1']=$matches[3];
                                $this->colours[$key]['blue1']=$matches[4];
                                $this->colours[$key]['bottom']=-2;
                                $this->colours[$key]['top']=-1;

                                $linematched++;
                            }

                            if (($last_seen == 'NODE') && (preg_match(
                                                               "/^\s*(LABELFONT|LABELFONTSHADOW|LABELBG|LABELOUTLINE)COLOR\s+(\d+)\s+(\d+)\s+(\d+)/i",
                                                               $buffer,
                                                               $matches)))
                            {
                                $key=$matches[1];
                                #   print "Found NODE colour line for $key\n";
                                $field=strtolower($matches[1]) . 'colour';
                                $curnode->$field=array
                                    (
                                        $matches[2],
                                        $matches[3],
                                        $matches[4]
                                    );

                                $linematched++;
                            }

                            if (($last_seen == 'LINK') && (preg_match(
                                                               "/^\s*(BWBOX|BWFONT|BWOUTLINE|OUTLINE)COLOR\s+(\d+)\s+(\d+)\s+(\d+)/i",
                                                               $buffer,
                                                               $matches)))
                            {
                                $key=$matches[1];
                                #    print "Found LINK colour line for $key\n";
                                $field=strtolower($matches[1]) . 'colour';
                                $curlink->$field=array
                                    (
                                        $matches[2],
                                        $matches[3],
                                        $matches[4]
                                    );

                                $linematched++;
                            }

                            if (($last_seen == 'NODE') && (preg_match(
                                                               "/^\s*(LABELFONTSHADOW|LABELBG|LABELOUTLINE)COLOR\s+none/i",
                                                               $buffer,
                                                               $matches)))
                            {
                                $key=$matches[1];
                                #     print "Found NODE non-colour line for $key\n";
                                $field=strtolower($matches[1]) . 'colour';
                                $curnode->$field=array
                                    (
                                        -1,
                                        -1,
                                        -1
                                    );

                                $linematched++;
                            }

                            if (($last_seen == 'LINK') && (preg_match(
                                                               "/^\s*(BWBOX|BWOUTLINE|OUTLINE)COLOR\s+none/i", $buffer,
                                                               $matches)))
                            {
                                $key=$matches[1];
                                #      print "Found LINK non-colour line for $key\n";
                                $field=strtolower($matches[1]) . 'colour';
                                $curlink->$field=array
                                    (
                                        -1,
                                        -1,
                                        -1
                                    );

                                $linematched++;
                            }

                            if (preg_match("/^\s*HTMLSTYLE\s+(static|overlib)/i", $buffer, $matches))
                            {
                                $this->htmlstyle=$matches[1];
                                $linematched++;
                            }

                            if ($last_seen == 'LINK' && preg_match(
                                                            "/^\s*ARROWSTYLE\s+(classic|compact)/i", $buffer, $matches))
                            {
                                $curlink->arrowstyle=$matches[1];
                                $linematched++;
                            }

                            if ($last_seen == 'LINK' && preg_match(
                                                            "/^\s*ARROWSTYLE\s+(\d+)\s+(\d+)/i", $buffer, $matches))
                            {
                                $curlink->arrowstyle=$matches[1] . ' ' . $matches[2];
                                $linematched++;
                            }

                            if ($last_seen == '---' && preg_match(
                                                           "/^\s*ARROWSTYLE\s+(classic|compact)/i", $buffer, $matches))
                            {
                                warn ("Global ARROWSTYLE is deprecated. Use LINK DEFAULT and ARROWSTYLE instead.\n");
                                $this->defaultlink->arrowstyle=$matches[1];
                                $linematched++;
                            }

                            if (preg_match("/^\s*TITLE\s+(.*)/i", $buffer, $matches))
                            {
                                $this->title=$matches[1];
                                $linematched++;
                            }

                            if (preg_match("/^\s*HTMLOUTPUTFILE\s+(.*)/i", $buffer, $matches))
                            {
                                $this->htmloutputfile=trim($matches[1]);
                                $linematched++;
                            }

                            if (preg_match("/^\s*IMAGEOUTPUTFILE\s+(.*)/i", $buffer, $matches))
                            {
                                $this->imageoutputfile=trim($matches[1]);
                                $linematched++;
                            }

                            if ($linematched == 0 && trim($buffer) != '') { warn
                                ("Unrecognised config on line $linecount: $buffer"); }

                            if ($linematched > 1) { warn
                                ("Same line ($linecount) interpreted twice. This is a program error. Please report to Howie with your config!\nThe line was: $buffer");
                                }
                        } // if blankline
                    }     // while

                    if ($last_seen == "NODE")
                    {
                        //   $this->nodes[$curnode->name] = $curnode;
                        if ($curnode->name == 'DEFAULT')
                        {
                            $this->defaultnode=$curnode;
                            debug ("Saving Default Node: " . $curnode->name . "\n");
                        }
                        else
                        {
                            $this->nodes[$curnode->name]=$curnode;
                            debug ("Saving Node: " . $curnode->name . "\n");
                        }
                    }

                    if ($last_seen == "LINK")
                    {
                        // $this->links[$curlink->name] = $curlink;
                        if ($curlink->name == 'DEFAULT')
                        {
                            $this->defaultlink=$curlink;
                            debug ("Saving Default Link: " . $curlink->name . "\n");
                        }
                        else
                        {
                            if (isset($curlink->a) && isset($curlink->b))
                            {
                                $this->links[$curlink->name]=$curlink;
                                debug ("Saving Link: " . $curlink->name . "\n");
                            }
                            else { warn ("Dropping LINK " . $curlink->name . " - it hasn't got 2 NODES!"); }
                        }
                    }
                } // if $fd
                else
                {
                    warn ("Couldn't open config file $filename for reading\n");
                    return (FALSE);
                }

                fclose ($fd);

                // load some default colouring, otherwise it all goes wrong
                if ($scalesseen == 0)
                {
                    debug ("Adding default SCALE set.\n");
                    $defaults=array
                        (
                            '1:10' => array('bottom' => 1, 'top' => 10, 'red1' => 140, 'green1' => 0, 'blue1' => 255),
                            '10:25' => array('bottom' => 10, 'top' => 25, 'red1' => 32, 'green1' => 32, 'blue1' => 255),
                            '25:40' => array('bottom' => 25, 'top' => 40, 'red1' => 0, 'green1' => 192, 'blue1' => 255),
                            '40:55' => array('bottom' => 40, 'top' => 55, 'red1' => 0, 'green1' => 240, 'blue1' => 0),
                            '55:70' => array('bottom' => 55, 'top' => 70, 'red1' => 240, 'green1' => 240, 'blue1' => 0),
                            '70:85' => array('bottom' => 70, 'top' => 85, 'red1' => 255, 'green1' => 192, 'blue1' => 0),
                            '85:100' => array('bottom' => 85, 'top' => 100, 'red1' => 255, 'green1' => 0, 'blue1' => 0)
                        );

                    foreach ($defaults as $key => $def)
                    {
                        $this->colours[$key]=$def;
                        $scalesseen++;
                    }
                }
                else { debug ("Already have $scalesseen scales, no defaults added.\n"); }

                $this->numscales=$scalesseen;
                $this->configfile="$filename";
                return (TRUE);
            }

            function WriteConfig($filename)
            {
                global $WEATHERMAP_VERSION;

                $fd=fopen($filename, "w");
                $output="";

                if ($fd)
                {
                    $output.="# Automatically generated by php-weathermap v$WEATHERMAP_VERSION\n\n";

                    if ($this->background != '') { $output.="BACKGROUND " . $this->background . "\n"; }
                    else
                    {
                        $output.="WIDTH " . $this->width . "\n";
                        $output.="HEIGHT " . $this->height . "\n";
                    }

                    if ($this->htmlstyle != $this->inherit_fieldlist['htmlstyle'])
                        { $output.="HTMLSTYLE " . $this->htmlstyle . "\n"; }

                    //      if( $this->keystyle != $this->inherit_fieldlist['keystyle']) { $output .= "KEYSTYLE ".$this->keystyle."\n"; }
                    if ($this->kilo != $this->inherit_fieldlist['kilo']) { $output.="KILO " . $this->kilo . "\n"; }

                    $output.="\n";

                    if (count($this->fonts) > 0)
                    {
                        foreach ($this->fonts as $fontnumber => $font)
                        {
                            if ($font->type == 'truetype')
                                $output.=sprintf("FONTDEFINE %d %s %d\n", $fontnumber, $font->file, $font->size);

                            if ($font->type == 'gd')
                                $output.=sprintf("FONTDEFINE %d %s\n", $fontnumber, $font->file);
                        }

                        $output.="\n";
                    }

                    if ($this->keyfont != $this->inherit_fieldlist['keyfont'])
                        { $output.="KEYFONT " . $this->keyfont . "\n"; }

                    if ($this->timefont != $this->inherit_fieldlist['timefont'])
                        { $output.="TIMEFONT " . $this->timefont . "\n"; }

                    if ($this->titlefont != $this->inherit_fieldlist['titlefont'])
                        { $output.="TITLEFONT " . $this->titlefont . "\n"; }

                    if (trim($this->title) != $this->inherit_fieldlist['title'])
                        { $output.="TITLE " . $this->title . "\n"; }

                    if (trim($this->htmlfile) != $this->inherit_fieldlist['htmlfile'])
                        { $output.="HTMLOUTPUTFILE " . $this->htmlfile . "\n"; }

                    if (trim($this->imagefile) != $this->inherit_fieldlist['imagefile'])
                        { $output.="IMAGEOUTPUTFILE " . $this->imagefile . "\n"; }

                    if (($this->keyx != $this->inherit_fieldlist['keyx'])
                        && ($this->keyy != $this->inherit_fieldlist['keyy']))
                        $output.="KEYPOS " . $this->keyx . " " . $this->keyy . " " . $this->keytext . "\n";

                    if (($this->timex != $this->inherit_fieldlist['timex'])
                        && ($this->timey != $this->inherit_fieldlist['timey']))
                        $output.="TIMEPOS " . $this->timex . " " . $this->timey . " " . $this->stamptext . "\n";

                    if (($this->titlex != $this->inherit_fieldlist['titlex'])
                        && ($this->titley != $this->inherit_fieldlist['titley']))
                        $output.="TITLEPOS " . $this->titlex . " " . $this->titley . "\n";

                    $output.="\n";

                    $colours=$this->colours;

                    #  fwrite($fd,"\n#     low  high   red green blue\n");
                    foreach ($colours as $k => $colour)
                    {
                        if ($colour['top'] >= 0)
                        {
                            if (!isset($colour['red2']))
                                $output.=sprintf("SCALE %d %d   %d %d %d\n", $colour['bottom'], $colour['top'],
                                                 $colour['red1'],            $colour['green1'], $colour['blue1']);
                            else
                                $output.=sprintf("SCALE %d %d   %d %d %d   %d %d %d\n", $colour['bottom'],
                                                 $colour['top'],                        $colour['red1'],
                                                 $colour['green1'],                     $colour['blue1'],
                                                 $colour['red2'],                       $colour['green2'],
                                                 $colour['blue2']);
                        }
                        else { $output.=sprintf("%sCOLOR %d %d %d\n", $k, $colour['red1'], $colour['green1'],
                                                $colour['blue1']); }
                    }

                    $output.="\n# End of global section\n\n# DEFAULT definitions:\n";

                    fwrite($fd, $output);

                    $this->defaultnode->WriteConfig($fd);
                    $this->defaultlink->WriteConfig($fd);

                    fwrite($fd, "\n# End of DEFAULTS section\n\n# Node definitions:\n");

                    foreach ($this->nodes as $node) { $node->WriteConfig($fd); }

                    fwrite($fd, "\n# End of NODE section\n\n# Link definitions:\n");

                    foreach ($this->links as $link) { $link->WriteConfig($fd); }

                    fwrite($fd, "\n# End of LINK section\n\n# That's All Folks!\n");
                }
                else
                {
                    warn ("Couldn't open config file $filename for writing");
                    return (FALSE);
                }

                return (TRUE);
            }

            // pre-allocate colour slots for the colours used by the arrows
            // this way, it's the pretty icons that suffer if there aren't enough colours, and
            // not the actual useful data
            // we skip any gradient scales
            function AllocateScaleColours($im)
            {
                $colours=$this->colours;

                foreach ($colours as $key => $colour)
                {
                    if (!isset($this->colours[$key]['red2']))
                    {
                        $r=$colour['red1'];
                        $g=$colour['green1'];
                        $b=$colour['blue1'];
                        debug ("AllocateScaleColours: $key ($r,$g,$b)\n");
                        $this->colours[$key]['gdref1']=myimagecolorallocate($im, $r, $g, $b);
                    }
                }
            }

            function DrawMap($filename = '', $thumbnailfile = '', $thumbnailmax = 250)
            {
                $bgimage=NULL;

                // do the basic prep work
                if ($this->background != '')
                {
                    if (is_readable($this->background))
                    {
                        $bgimage=imagecreatefrompng($this->background);

                        if (!$bgimage) { warn
                            ("Failed to open background image.  One possible reason: Is your BACKGROUND really a PNG?\n");
                            }
                        else
                        {
                            $this->width=imagesx($bgimage);
                            $this->height=imagesy($bgimage);
                        }
                    }
                    else { warn
                        ("Your background image file could not be read. Check the filename, and permissions, for "
                            . $this->background . "\n"); }
                }

                $image=imagecreatetruecolor($this->width, $this->height);

                //        $image = imagecreate($this->width, $this->height);
                if (!$image) { warn
                    ("Couldn't create output image in memory (" . $this->width . "x" . $this->height . ")."); }
                else
                {
                    ImageAlphaBlending($image, true);
                    //imageantialias($image,true);

                    // by here, we should have a valid image handle

                    // save this away, now
                    $this->image=$image;

                    $this->white=myimagecolorallocate($image, 255, 255, 255);
                    $this->black=myimagecolorallocate($image, 0, 0, 0);
                    $this->grey=myimagecolorallocate($image, 192, 192, 192);
                    $this->selected=myimagecolorallocate($image, 255, 0, 0); // for selections in the editor

                    $this->AllocateScaleColours($image);

                    // fill with background colour anyway, in case the background image failed to load
                    imagefilledrectangle($image, 0, 0, $this->width, $this->height, $this->colours['BG']['gdref1']);

                    if ($bgimage)
                    {
                        imagecopy($image, $bgimage, 0, 0, 0, 0, $this->width, $this->height);
                        imagedestroy ($bgimage);
                    }

                    // Now it's time to draw a map

                    if ($this->need_size_precalc)
                    {
                        foreach ($this->nodes as $node) { $this->nodes[$node->name]->calc_size(); }
                    }

                    foreach ($this->links as $link) { $link->Draw($image, $this); }

                    foreach ($this->nodes as $node) { $node->Draw($image, $this); }

                    $this->DrawLegend_Classic($image);

                    $this->DrawTimestamp($image, $this->timefont, $this->colours['TIME']['gdref1']);
                    $this->DrawTitle($image, $this->titlefont, $this->colours['TITLE']['gdref1']);

                    // Ready to output the results...

                    if ($filename == '') { imagepng ($image); }
                    elseif ($filename == 'null') {
                    // do nothing at all - we just wanted the HTML AREAs for the editor or HTML output
                        }
                    else { imagepng($image, $filename); }

                    if (function_exists('imagecopyresampled'))
                    {
                        // if one is specified, and we can, write a thumbnail too
                        if ($thumbnailfile != '')
                        {
                            if ($this->width > $this->height) { $factor=($thumbnailmax / $this->width); }
                            else { $factor=($thumbnailmax / $this->height); }
    
                            $twidth=$this->width * $factor;
                            $theight=$this->height * $factor;
    
                            $imagethumb=imagecreatetruecolor($twidth, $theight);
                            imagecopyresampled($imagethumb, $image, 0, 0, 0, 0, $twidth, $theight, $this->width,
                                               $this->height);
                            imagepng($imagethumb, $thumbnailfile);
                        }
                    }
                    else
                    {
                            warn("Skipping thumbnail creation, since we don't have the necessary function.");
                    }

                    imagedestroy ($image);
                }
            }

            function PreloadMapHTML()
            {
                if ($this->htmlstyle == "overlib")
                {
                    #   onmouseover="return overlib('<img src=graph.png>',DELAY,250,CAPTION,'$caption');"  onmouseout="return nd();"

                    $center_x=$this->width / 2;
                    $center_y=$this->height / 2;

                    foreach ($this->links as $link)
                    {
                        if ($link->overliburl != '')
                        {
                            # $overlibhtml = "onmouseover=\"return overlib('&lt;img src=".$link->overliburl."&gt;',DELAY,250,CAPTION,'".$link->name."');\"  onmouseout=\"return nd();\"";
                            $a_x=$link->a->x;
                            $b_x=$link->b->x;
                            $mid_x=($a_x + $b_x) / 2;
                            $a_y=$link->a->y;
                            $b_y=$link->b->y;
                            $mid_y=($a_y + $b_y) / 2;

                            //            debug($link->overlibwidth."---".$link->overlibheight."---\n");

                            $left="";
                            $above="";

                            if ($link->overlibwidth != 0)
                            {
                                $left="WIDTH," . $link->overlibwidth . ",";

                                if ($mid_x > $center_x)
                                    $left.="LEFT,";
                            }

                            if ($link->overlibheight != 0)
                            {
                                $above="HEIGHT," . $link->overlibheight . ",";

                                if ($mid_y > $center_y)
                                    $above.="ABOVE,";
                            }

                            $overlibhtml="onmouseover=\"return overlib('&lt;img src=" . $link->overliburl
                                             . "&gt;',DELAY,250,${left}${above}CAPTION,'" . $link->name
                                             . "');\"  onmouseout=\"return nd();\"";

                            $this->imap->setProp("extrahtml", $overlibhtml, "LINK:" . $link->name);
                        }
                    }

                    foreach ($this->nodes as $node)
                    {
                        if ($node->overliburl != '')
                        {
                            # $overlibhtml = "onmouseover=\"return overlib('&lt;img src=".$node->overliburl."&gt;',DELAY,250,CAPTION,'".$node->name."');\"  onmouseout=\"return nd();\"";

                            debug ($node->overlibwidth . "---" . $node->overlibheight . "---\n");

                            $left="";
                            $above="";

                            if ($node->overlibwidth != 0)
                            {
                                $left="WIDTH," . $node->overlibwidth . ",";

                                if ($node->x > $center_x)
                                    $left.="LEFT,";
                            }

                            if ($node->overlibheight != 0)
                            {
                                $above="HEIGHT," . $node->overlibheight . ",";

                                if ($node->y > $center_y)
                                    $above.="ABOVE,";
                            }

                            $overlibhtml="onmouseover=\"return overlib('&lt;img src=" . $node->overliburl
                                             . "&gt;',DELAY,250,${left}${above}CAPTION,'" . $node->name
                                             . "');\"  onmouseout=\"return nd();\"";

                            $this->imap->setProp("extrahtml", $overlibhtml, "NODE:" . $node->name);
                        }
                    }
                }

                if ($this->htmlstyle == 'editor')
                {
                    foreach ($this->links as $link) {
                    //   $this->imap->setProp("href","#","LINK:".$link->name);
                    //  $this->imap->setProp("extrahtml","onclick=\"click_handler('link','".$link->name."');\"","LINK:".$link->name);
                        }

                    foreach ($this->nodes as $node) {
                    //    $this->imap->setProp("href","#","NODE:".$node->name);
                    //    $this->imap->setProp("extrahtml","onclick=\"click_handler('node','".$node->name."');\"","NODE:".$node->name);
                    # $this->imap->setProp("extrahtml","onclick=\"alert('".$node->name."');\"","NODE:".$node->name);
                        }
                }
                else
                {
                    foreach ($this->links as $link)
                    {
                        if ($link->infourl != '') { $this->imap->setProp("href", $link->infourl,
                                                                         "LINK:" . $link->name); }
                    }

                    foreach ($this->nodes as $node)
                    {
                        if ($node->infourl != '') { $this->imap->setProp("href", $node->infourl,
                                                                         "NODE:" . $node->name); }
                    }
                }
            }

            function asJS()
            {
                $js='';

                $js.="var Links = new Array();\n";
                $js.=$this->defaultlink->asJS();

                foreach ($this->links as $link) { $js.=$link->asJS(); }

                $js.="var Nodes = new Array();\n";
                $js.=$this->defaultnode->asJS();

                foreach ($this->nodes as $node) { $js.=$node->asJS(); }

                return $js;
            }

            function asJSON()
            {
                $json = '';
                
                $json .= "{ \n";
                                
                $json .= "'map': {  \n";
                foreach (array_keys($this->inherit_fieldlist)as $fld)
                {
                    $json .= js_escape($fld).": ";
                    $json .= js_escape($this->$fld);
                    $json .= ",\n";
                }
                $json = rtrim($json,", \n");
                $json .= "\n},\n";
                                
                $json .= "'nodes': {\n";
                $json .= $this->defaultnode->asJSON();
                foreach ($this->nodes as $node) { $json .= $node->asJSON(); }
                $json = rtrim($json,", \n");
                $json .= "\n},\n";
                
                
                
                $json .= "'links': {\n";
                $json .= $this->defaultlink->asJSON();
                foreach ($this->links as $link) { $json .= $link->asJSON(); }
                $json = rtrim($json,", \n");
                $json .= "\n},\n";
                
                $json .= "'imap': [\n";
                $json .= $this->imap->subJSON("NODE:");
                // should check if there WERE nodes...
                $json .= ",\n";
                $json .= $this->imap->subJSON("LINK:");
                $json .= "\n]\n";
                $json .= "\n";
                
                $json .= ", valid: 1}\n";
                
                return($json);
            }

            // imagemapname is a parameter, so we can stack up several maps in the Cacti plugin
            function MakeHTML($imagemapname = "weathermap_imap")
            {
                $this->PreloadMapHTML();
                $html='';
				
                if ($this->imageuri != '') { 
                	$matches = preg_split('/\//', $this->imageuri);
					$i = -1;
					foreach ($matches as $key)
        				$i++;
                	$html.=sprintf('<img src="%s" width="%s" height="%s" border="0" usemap="#'. $imagemapname.'" alt="network weathermap"/>',
                                   $matches[$i], $this->width, $this->height); 
                } else { 
                	$matches = preg_split('/\//', $this->imagefile);
					$i = -1;
					foreach ($matches as $key)
	        			$i++;
                	$html.=sprintf('<img src="%s" width="%s" height="%s" border="0" usemap="#'.$imagemapname.'" alt="network weathermap" />',
                                   $matches[$i], $this->width,$this->height); }

                $html.='<map name="' . $imagemapname . '" id="' . $imagemapname . '">';

                $html.=$this->imap->subHTML("NODE:");
                $html.=$this->imap->subHTML("LINK:",true);

                $html.='</map>';

                return ($html);
            }
        }

    ;
?>
